# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
"""
The cyipopt_interface module includes the python interface to the
Cythonized ipopt solver cyipopt (see more:
https://github.com/mechmotum/cyipopt.git). To use the interface,
you can create a derived implementation from the abstract base class
CyIpoptProblemInterface that provides the necessary methods.

Note: This module also includes a default implementation CyIpopt
that works with problems derived from AslNLP as long as those
classes return numpy ndarray objects for the vectors and coo_matrix
objects for the matrices (e.g., AmplNLP and PyomoNLP)
"""

import abc
import inspect

from pyomo.common.dependencies import attempt_import, numpy as np, numpy_available
from pyomo.contrib.pynumero.exceptions import PyNumeroEvaluationError


def _cyipopt_importer():
    import cyipopt

    # cyipopt before version 1.0.3 called the problem class "Problem"
    if not hasattr(cyipopt, "Problem"):
        cyipopt.Problem = cyipopt.problem
    # cyipopt before version 1.0.3 put the __version__ flag in the ipopt
    # module (which was deprecated starting in 1.0.3)
    if not hasattr(cyipopt, "__version__"):
        import ipopt

        cyipopt.__version__ = ipopt.__version__
    # Beginning in 1.0.3, STATUS_MESSAGES is in a separate
    # ipopt_wrapper module
    if not hasattr(cyipopt, "STATUS_MESSAGES"):
        import ipopt_wrapper

        cyipopt.STATUS_MESSAGES = ipopt_wrapper.STATUS_MESSAGES
    return cyipopt


cyipopt, cyipopt_available = attempt_import(
    "ipopt",
    error_message="cyipopt solver relies on the ipopt module from cyipopt. "
    "See https://github.com/mechmotum/cyipopt.git for cyipopt "
    "installation instructions.",
    importer=_cyipopt_importer,
)

# If cyipopt is not available, we will use object as our base class for
# CyIpoptProblemInterface so we don't require cyipopt to import from
# this file.
# Note that this *does* trigger the import attempt and therefore is
# moderately time-consuming.
cyipopt_Problem = cyipopt.Problem if cyipopt_available else object


class CyIpoptProblemInterface(cyipopt_Problem, metaclass=abc.ABCMeta):
    """Abstract subclass of ``cyipopt.Problem`` defining an object that can be
    used as an interface to CyIpopt. Subclasses must define all methods necessary
    for the CyIpopt solve and must call this class's ``__init__`` method to
    initialize Ipopt's data structures.

    Note that, if "output_file" is provided as an Ipopt option, the log file
    is open until this object (and thus the underlying Ipopt NLP object) is
    deallocated. To force this deallocation, call the ``close()`` method, which
    is defined by ``cyipopt.Problem``.

    """

    # Flag used to determine whether the underlying IpoptProblem struct
    # has been initialized. This is used to prevent segfaults when calling
    # cyipopt.Problem's solve method if cyipopt.Problem.__init__ hasn't been
    # called.
    _problem_initialized = False

    def __init__(self):
        """Initialize the problem interface

        This method calls ``cyipopt.Problem.__init__``, and *must* be called
        by any subclass's ``__init__`` method. If not, we will segfault when
        we call ``cyipopt.Problem.solve`` from this object.

        """
        if not cyipopt_available:
            raise RuntimeError(
                "cyipopt is required to instantiate CyIpoptProblemInterface"
            )

        # Call cyipopt.Problem.__init__
        xl = self.x_lb()
        xu = self.x_ub()
        gl = self.g_lb()
        gu = self.g_ub()
        nx = len(xl)
        ng = len(gl)
        super(CyIpoptProblemInterface, self).__init__(
            n=nx, m=ng, lb=xl, ub=xu, cl=gl, cu=gu
        )
        # Set a flag to indicate that the IpoptProblem struct has been
        # initialized
        self._problem_initialized = True

    def solve(self, x, lagrange=None, zl=None, zu=None):
        """Solve a CyIpopt Problem

        Checks whether __init__ has been called before calling
        cyipopt.Problem.solve

        """
        lagrange = [] if lagrange is None else lagrange
        zl = [] if zl is None else zl
        zu = [] if zu is None else zu
        # Check a flag to make sure __init__ has been called. This is to prevent
        # segfaults if we try to call solve from a subclass that has not called
        # super().__init__
        #
        # Note that we can still segfault if a user overrides solve and does not
        # call cyipopt.Problem.__init__, but in this case we assume they know what
        # they are doing.
        if not self._problem_initialized:
            raise RuntimeError(
                "Attempting to call cyipopt.Problem.solve when"
                " cyipopt.Problem.__init__ has not been called. This can happen"
                " if a subclass of CyIpoptProblemInterface overrides __init__"
                " without calling CyIpoptProblemInterface.__init__ or setting"
                " the CyIpoptProblemInterface._problem_initialized flag."
            )
        return super(CyIpoptProblemInterface, self).solve(
            x, lagrange=lagrange, zl=zl, zu=zu
        )

    @abc.abstractmethod
    def x_init(self):
        """Return the initial values for x as a numpy ndarray"""
        pass

    @abc.abstractmethod
    def x_lb(self):
        """Return the lower bounds on x as a numpy ndarray"""
        pass

    @abc.abstractmethod
    def x_ub(self):
        """Return the upper bounds on x as a numpy ndarray"""
        pass

    @abc.abstractmethod
    def g_lb(self):
        """Return the lower bounds on the constraints as a numpy ndarray"""
        pass

    @abc.abstractmethod
    def g_ub(self):
        """Return the upper bounds on the constraints as a numpy ndarray"""
        pass

    @abc.abstractmethod
    def scaling_factors(self):
        """Return the values for scaling factors as a tuple
        (objective_scaling, x_scaling, g_scaling). Return None
        if the scaling factors are to be ignored
        """
        pass

    @abc.abstractmethod
    def objective(self, x):
        """Return the value of the objective
        function evaluated at x
        """
        pass

    @abc.abstractmethod
    def gradient(self, x):
        """Return the gradient of the objective
        function evaluated at x as a numpy ndarray
        """
        pass

    @abc.abstractmethod
    def constraints(self, x):
        """Return the residuals of the constraints
        evaluated at x as a numpy ndarray
        """
        pass

    @abc.abstractmethod
    def jacobianstructure(self):
        """Return the structure of the jacobian
        in coordinate format. That is, return (rows,cols)
        where rows and cols are both numpy ndarray
        objects that contain the row and column indices
        for each of the nonzeros in the jacobian.
        """
        pass

    @abc.abstractmethod
    def jacobian(self, x):
        """Return the values for the jacobian evaluated at x
        as a numpy ndarray of nonzero values corresponding
        to the rows and columns specified in the jacobianstructure
        """
        pass

    @abc.abstractmethod
    def hessianstructure(self):
        """Return the structure of the hessian
        in coordinate format. That is, return (rows,cols)
        where rows and cols are both numpy ndarray
        objects that contain the row and column indices
        for each of the nonzeros in the hessian.
        Note: return ONLY the lower diagonal of this symmetric matrix.
        """
        pass

    @abc.abstractmethod
    def hessian(self, x, y, obj_factor):
        """Return the values for the hessian evaluated at x
        as a numpy ndarray of nonzero values corresponding
        to the rows and columns specified in the
        hessianstructure method.
        Note: return ONLY the lower diagonal of this symmetric matrix.
        """
        pass

    def intermediate(
        self,
        alg_mod,
        iter_count,
        obj_value,
        inf_pr,
        inf_du,
        mu,
        d_norm,
        regularization_size,
        alpha_du,
        alpha_pr,
        ls_trials,
    ):
        """Callback that can be used to examine or report intermediate
        results. This method is called each iteration
        """
        # TODO: Document the arguments
        pass


class CyIpoptNLP(CyIpoptProblemInterface):
    def __init__(self, nlp, intermediate_callback=None, halt_on_evaluation_error=None):
        """This class provides a CyIpoptProblemInterface for use
        with the CyIpoptSolver class that can take in an NLP
        as long as it provides vectors as numpy ndarrays and
        matrices as scipy.sparse.coo_matrix objects. This class
        provides the interface between AmplNLP or PyomoNLP objects
        and the CyIpoptSolver
        """
        self._nlp = nlp
        self._intermediate_callback = intermediate_callback

        cyipopt_has_eval_error = cyipopt_available and hasattr(
            cyipopt, "CyIpoptEvaluationError"
        )
        if halt_on_evaluation_error is None:
            # If using cyipopt >= 1.3, the default is to continue.
            # Otherwise, the default is to halt (because we are forced to).
            #
            # If CyIpopt is not available, we "halt" (re-raise the original
            # exception).
            self._halt_on_evaluation_error = not cyipopt_has_eval_error
        elif not halt_on_evaluation_error and not cyipopt_has_eval_error:
            raise ValueError(
                "halt_on_evaluation_error=False is only supported for cyipopt >= 1.3.0"
            )
        else:
            self._halt_on_evaluation_error = halt_on_evaluation_error

        x = nlp.init_primals()
        y = nlp.init_duals()
        if np.any(np.isnan(y)):
            # did not get initial values for y, use this default
            y.fill(1.0)

        self._cached_x = x.copy()
        self._cached_y = y.copy()
        self._cached_obj_factor = 1.0

        nlp.set_primals(self._cached_x)
        nlp.set_duals(self._cached_y)

        # get jacobian and hessian structures
        self._jac_g = nlp.evaluate_jacobian()
        try:
            self._hess_lag = nlp.evaluate_hessian_lag()
            self._hess_lower_mask = self._hess_lag.row >= self._hess_lag.col
            self._hessian_available = True
        except (AttributeError, NotImplementedError):
            self._hessian_available = False
            self._hess_lag = None
            self._hess_lower_mask = None

        # Call CyIpoptProblemInterface.__init__, which calls
        # cyipopt.Problem.__init__
        super(CyIpoptNLP, self).__init__()

        # Pre-Pyomo 6.8.0, we had no way to pass the cyipopt.Problem object
        # to the user in an intermediate callback. This prevented them from calling
        # the useful get_current_iterate and get_current_violations methods. Now,
        # we support this by adding the Problem object to the args we pass to a user's
        # callback. To preserve backwards compatibility, we inspect the user's
        # callback to infer whether they want this argument. To preserve backwards
        # compatibility if the user asked for variable-length *args, we do not pass
        # the Problem object as an argument in this case.
        # A more maintainable solution may be to force users to accept **kwds if they
        # want "extra info." If we find ourselves continuing to augment this callback,
        # this may be worth considering. -RBP
        self._use_13arg_callback = None
        if self._intermediate_callback is not None:
            signature = inspect.signature(self._intermediate_callback)
            positional_kinds = {
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                inspect.Parameter.POSITIONAL_ONLY,
            }
            positional = [
                param
                for param in signature.parameters.values()
                if param.kind in positional_kinds
            ]
            has_var_args = any(
                p.kind is inspect.Parameter.VAR_POSITIONAL
                for p in signature.parameters.values()
            )

            if len(positional) == 13 and not has_var_args:
                # If *args is expected, we do not use the new callback
                # signature.
                self._use_13arg_callback = True
            elif len(positional) == 12 or has_var_args:
                # If *args is expected, we use the old callback signature
                # for backwards compatibility.
                self._use_13arg_callback = False
            else:
                raise ValueError(
                    "Invalid intermediate callback. A function with either 12 or 13"
                    " positional arguments, or a variable number of arguments, is"
                    " expected."
                )

    def _set_primals_if_necessary(self, x):
        if not np.array_equal(x, self._cached_x):
            self._nlp.set_primals(x)
            self._cached_x = x.copy()

    def _set_duals_if_necessary(self, y):
        if not np.array_equal(y, self._cached_y):
            self._nlp.set_duals(y)
            self._cached_y = y.copy()

    def _set_obj_factor_if_necessary(self, obj_factor):
        if obj_factor != self._cached_obj_factor:
            self._nlp.set_obj_factor(obj_factor)
            self._cached_obj_factor = obj_factor

    def x_init(self):
        return self._nlp.init_primals()

    def x_lb(self):
        return self._nlp.primals_lb()

    def x_ub(self):
        return self._nlp.primals_ub()

    def g_lb(self):
        return self._nlp.constraints_lb()

    def g_ub(self):
        return self._nlp.constraints_ub()

    def scaling_factors(self):
        obj_scaling = self._nlp.get_obj_scaling()
        x_scaling = self._nlp.get_primals_scaling()
        g_scaling = self._nlp.get_constraints_scaling()
        return obj_scaling, x_scaling, g_scaling

    def objective(self, x):
        try:
            self._set_primals_if_necessary(x)
            return self._nlp.evaluate_objective()
        except PyNumeroEvaluationError:
            if self._halt_on_evaluation_error:
                raise
            else:
                raise cyipopt.CyIpoptEvaluationError(
                    "Error in objective function evaluation"
                )

    def gradient(self, x):
        try:
            self._set_primals_if_necessary(x)
            return self._nlp.evaluate_grad_objective()
        except PyNumeroEvaluationError:
            if self._halt_on_evaluation_error:
                raise
            else:
                raise cyipopt.CyIpoptEvaluationError(
                    "Error in objective gradient evaluation"
                )

    def constraints(self, x):
        try:
            self._set_primals_if_necessary(x)
            return self._nlp.evaluate_constraints()
        except PyNumeroEvaluationError:
            if self._halt_on_evaluation_error:
                raise
            else:
                raise cyipopt.CyIpoptEvaluationError("Error in constraint evaluation")

    def jacobianstructure(self):
        return self._jac_g.row, self._jac_g.col

    def jacobian(self, x):
        try:
            self._set_primals_if_necessary(x)
            self._nlp.evaluate_jacobian(out=self._jac_g)
            return self._jac_g.data
        except PyNumeroEvaluationError:
            if self._halt_on_evaluation_error:
                raise
            else:
                raise cyipopt.CyIpoptEvaluationError(
                    "Error in constraint Jacobian evaluation"
                )

    def hessianstructure(self):
        if not self._hessian_available:
            return np.zeros(0), np.zeros(0)

        row = np.compress(self._hess_lower_mask, self._hess_lag.row)
        col = np.compress(self._hess_lower_mask, self._hess_lag.col)
        return row, col

    def hessian(self, x, y, obj_factor):
        if not self._hessian_available:
            raise ValueError("Hessian requested, but not supported by the NLP")

        try:
            self._set_primals_if_necessary(x)
            self._set_duals_if_necessary(y)
            self._set_obj_factor_if_necessary(obj_factor)
            self._nlp.evaluate_hessian_lag(out=self._hess_lag)
            data = np.compress(self._hess_lower_mask, self._hess_lag.data)
            return data
        except PyNumeroEvaluationError:
            if self._halt_on_evaluation_error:
                raise
            else:
                raise cyipopt.CyIpoptEvaluationError(
                    "Error in Lagrangian Hessian evaluation"
                )

    def intermediate(
        self,
        alg_mod,
        iter_count,
        obj_value,
        inf_pr,
        inf_du,
        mu,
        d_norm,
        regularization_size,
        alpha_du,
        alpha_pr,
        ls_trials,
    ):
        """Calls user's intermediate callback

        This method has the call signature expected by CyIpopt. We then extend
        this call signature to provide users of this interface class additional
        functionality. Additional arguments are:

        - The ``NLP`` object that was used to construct this class instance.
          This is useful for querying the variables, constraints, and
          derivatives during the callback.
        - The class instance itself. This is useful for calling the
          ``get_current_iterate`` and ``get_current_violations`` methods, which
          query Ipopt's internal data structures to provide this information.

        """
        if self._intermediate_callback is not None:
            if self._use_13arg_callback:
                # This is the callback signature expected as of Pyomo 6.8.0
                return self._intermediate_callback(
                    self._nlp,
                    self,
                    alg_mod,
                    iter_count,
                    obj_value,
                    inf_pr,
                    inf_du,
                    mu,
                    d_norm,
                    regularization_size,
                    alpha_du,
                    alpha_pr,
                    ls_trials,
                )
            else:
                # This is the callback signature expected pre-Pyomo 6.8.0 and
                # is supported for backwards compatibility.
                return self._intermediate_callback(
                    self._nlp,
                    alg_mod,
                    iter_count,
                    obj_value,
                    inf_pr,
                    inf_du,
                    mu,
                    d_norm,
                    regularization_size,
                    alpha_du,
                    alpha_pr,
                    ls_trials,
                )
        return True



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pickle

import pyomo.common.unittest as unittest
import pyomo.kernel as pmo
from pyomo.core.kernel.base import ICategorizedObject, ICategorizedObjectContainer
from pyomo.core.kernel.homogeneous_container import IHomogeneousContainer
from pyomo.core.kernel.tuple_container import TupleContainer
from pyomo.core.kernel.constraint import (
    IConstraint,
    constraint,
    constraint_dict,
    constraint_tuple,
    constraint_list,
)
from pyomo.core.kernel.matrix_constraint import matrix_constraint, _MatrixConstraintData
from pyomo.core.kernel.variable import variable, variable_list
from pyomo.core.kernel.parameter import parameter
from pyomo.core.kernel.expression import expression
from pyomo.core.kernel.block import block, block_list

try:
    import numpy

    has_numpy = True
except:
    has_numpy = False

try:
    import scipy

    has_scipy = True
    _scipy_ver = tuple(int(_) for _ in scipy.version.version.split('.')[:2])
except:
    has_scipy = False
    _scipy_ver = (0, 0)


def _create_variable_list(size, **kwds):
    assert size > 0
    vlist = variable_list()
    for i in range(size):
        vlist.append(variable(**kwds))
    return vlist


@unittest.skipUnless(has_numpy and has_scipy, "NumPy or SciPy is not available")
class Test_matrix_constraint(unittest.TestCase):
    def test_pprint(self):
        # Not really testing what the output is, just that
        # an error does not occur. The pprint functionality
        # is still in the early stages.
        m, n = 3, 2
        vlist = _create_variable_list(2)
        A = numpy.random.rand(m, n)
        ctuple = matrix_constraint(A, lb=1, ub=2, x=vlist)
        pmo.pprint(ctuple)
        b = block()
        b.c = ctuple
        pmo.pprint(ctuple)
        pmo.pprint(b)
        m = block()
        m.b = b
        pmo.pprint(ctuple)
        pmo.pprint(b)
        pmo.pprint(m)

    def test_ctype(self):
        ctuple = matrix_constraint(numpy.random.rand(3, 3))
        self.assertIs(ctuple.ctype, IConstraint)
        self.assertIs(type(ctuple), matrix_constraint)
        self.assertIs(type(ctuple)._ctype, IConstraint)
        self.assertIs(ctuple[0].ctype, IConstraint)
        self.assertIs(type(ctuple[0])._ctype, IConstraint)

    def test_pickle(self):
        vlist = _create_variable_list(3)
        ctuple = matrix_constraint(numpy.array([[0, 0, 0], [0, 0, 0]]), x=vlist)
        self.assertTrue((ctuple.lb == -numpy.inf).all())
        self.assertTrue((ctuple.ub == numpy.inf).all())
        self.assertTrue((ctuple.equality == False).all())
        self.assertEqual(ctuple.parent, None)
        ctuple_up = pickle.loads(pickle.dumps(ctuple))
        self.assertTrue((ctuple_up.lb == -numpy.inf).all())
        self.assertTrue((ctuple_up.ub == numpy.inf).all())
        self.assertTrue((ctuple_up.equality == False).all())
        self.assertEqual(ctuple_up.parent, None)
        b = block()
        b.ctuple = ctuple
        self.assertIs(ctuple.parent, b)
        bup = pickle.loads(pickle.dumps(b))
        ctuple_up = bup.ctuple
        self.assertTrue((ctuple_up.lb == -numpy.inf).all())
        self.assertTrue((ctuple_up.ub == numpy.inf).all())
        self.assertTrue((ctuple_up.equality == False).all())
        self.assertIs(ctuple_up.parent, bup)

    def test_init(self):
        vlist = _create_variable_list(3, value=1.0)
        ctuple = matrix_constraint(numpy.zeros((3, 3)), x=vlist)
        self.assertEqual(len(ctuple), 3)
        self.assertTrue(ctuple.parent is None)
        self.assertEqual(ctuple.ctype, IConstraint)
        self.assertTrue((ctuple.lb == -numpy.inf).all())
        self.assertTrue((ctuple.ub == numpy.inf).all())
        self.assertTrue((ctuple.equality == False).all())
        for c in ctuple:
            self.assertEqual(c.lb, None)
            self.assertEqual(c.ub, None)
            self.assertEqual(c.equality, False)
            self.assertEqual(c(), 0)
            self.assertEqual(c.slack, float('inf'))
            self.assertEqual(c.lslack, float('inf'))
            self.assertEqual(c.uslack, float('inf'))
            self.assertEqual(c.has_lb(), False)
            self.assertEqual(c.has_ub(), False)

        vlist = _create_variable_list(3, value=3)
        A = numpy.ones((2, 3))
        ctuple = matrix_constraint(A, lb=0, ub=2, x=vlist)
        self.assertEqual(len(ctuple), 2)
        self.assertTrue((ctuple.lb == 0).all())
        self.assertTrue((ctuple.ub == 2).all())
        self.assertTrue((ctuple.equality == False).all())
        for c in ctuple:
            self.assertEqual(len(list(c.terms)), 3)
            self.assertEqual(c.lb, 0)
            self.assertEqual(c.body(), 9)
            self.assertEqual(c(), 9)
            self.assertEqual(c.ub, 2)

        ctuple = matrix_constraint(A, rhs=1, x=vlist)
        self.assertEqual(len(ctuple), 2)
        self.assertTrue((ctuple.lb == 1).all())
        self.assertTrue((ctuple.ub == 1).all())
        self.assertTrue((ctuple.rhs == 1).all())
        self.assertTrue((ctuple.equality == True).all())
        for c in ctuple:
            self.assertEqual(len(list(c.terms)), 3)
            self.assertEqual(c.lb, 1)
            self.assertEqual(c.body(), 9)
            self.assertEqual(c(), 9)
            self.assertEqual(c.ub, 1)
            self.assertEqual(c.rhs, 1)

        # can't use both lb and rhs
        with self.assertRaises(ValueError):
            matrix_constraint(A, lb=0, rhs=0, x=vlist)
        # can't use both ub and rhs
        with self.assertRaises(ValueError):
            matrix_constraint(A, ub=0, rhs=0, x=vlist)

    def test_type(self):
        A = numpy.ones((2, 3))
        ctuple = matrix_constraint(A)
        self.assertTrue(isinstance(ctuple, ICategorizedObject))
        self.assertTrue(isinstance(ctuple, ICategorizedObjectContainer))
        self.assertTrue(isinstance(ctuple, IHomogeneousContainer))
        self.assertTrue(isinstance(ctuple, TupleContainer))
        self.assertTrue(isinstance(ctuple, constraint_tuple))
        self.assertTrue(isinstance(ctuple, matrix_constraint))
        self.assertTrue(isinstance(ctuple[0], ICategorizedObject))
        self.assertTrue(isinstance(ctuple[0], IConstraint))
        self.assertTrue(isinstance(ctuple[0], _MatrixConstraintData))

    def test_active(self):
        A = numpy.ones((2, 2))
        ctuple = matrix_constraint(A)
        self.assertEqual(ctuple.active, True)
        for c in ctuple:
            self.assertEqual(c.active, True)
        ctuple.deactivate()
        self.assertEqual(ctuple.active, False)
        for c in ctuple:
            self.assertEqual(c.active, True)
        ctuple.deactivate(shallow=False)
        self.assertEqual(ctuple.active, False)
        for c in ctuple:
            self.assertEqual(c.active, False)
        ctuple[0].activate()
        self.assertEqual(ctuple.active, False)
        self.assertEqual(ctuple[0].active, True)
        self.assertEqual(ctuple[1].active, False)
        ctuple.activate()
        self.assertEqual(ctuple.active, True)
        self.assertEqual(ctuple[0].active, True)
        self.assertEqual(ctuple[1].active, False)
        ctuple.activate(shallow=False)
        self.assertEqual(ctuple.active, True)
        for c in ctuple:
            self.assertEqual(c.active, True)

        b = block()
        self.assertEqual(b.active, True)
        b.deactivate()
        self.assertEqual(b.active, False)
        b.c = ctuple
        self.assertEqual(ctuple.active, True)
        self.assertEqual(b.active, False)
        ctuple.deactivate()
        self.assertEqual(ctuple.active, False)
        self.assertEqual(b.active, False)
        b.activate()
        self.assertEqual(ctuple.active, False)
        self.assertEqual(b.active, True)
        b.activate(shallow=False)
        self.assertEqual(ctuple.active, True)
        self.assertEqual(b.active, True)
        b.deactivate(shallow=False)
        self.assertEqual(ctuple.active, False)
        self.assertEqual(b.active, False)

    def test_index(self):
        A = numpy.ones((4, 5))
        ctuple = matrix_constraint(A)
        for i, c in enumerate(ctuple):
            self.assertEqual(c.index, i)

    @unittest.skipIf(
        _scipy_ver < (1, 1), "csr_matrix.reshape only available in scipy >= 1.1"
    )
    def test_A(self):
        A = numpy.ones((4, 5))

        # sparse
        c = matrix_constraint(A)
        self.assertEqual(c.A.shape, A.shape)
        self.assertTrue((c.A == A).all())
        self.assertEqual(c.sparse, True)
        with self.assertRaises(ValueError):
            c.A.data[0] = 2
        with self.assertRaises(ValueError):
            c.A.indices[0] = 2
        with self.assertRaises(ValueError):
            c.A.indptr[0] = 2
        cA = c.A
        cA.shape = (5, 4)
        # the shape of c.A should not be changed
        self.assertEqual(c.A.shape, (4, 5))

        # dense
        c = matrix_constraint(A, sparse=False)
        self.assertEqual(c.A.shape, A.shape)
        self.assertTrue((c.A == A).all())
        self.assertEqual(c.sparse, False)
        with self.assertRaises(ValueError):
            c.A[0, 0] = 2
        cA = c.A
        cA.shape = (5, 4)
        # the shape of c.A should not be changed
        self.assertEqual(c.A.shape, (4, 5))

    def test_x(self):
        A = numpy.ones((4, 5))
        ctuple = matrix_constraint(A)
        self.assertEqual(ctuple.x, None)
        for c in ctuple:
            with self.assertRaises(ValueError):
                list(c.terms)
        vlist = _create_variable_list(5)
        ctuple.x = vlist
        self.assertEqual(len(ctuple.x), 5)
        self.assertEqual(len(ctuple.x), len(vlist))
        self.assertIsNot(ctuple.x, len(vlist))
        for i, v in enumerate(ctuple.x):
            self.assertIs(v, vlist[i])

        ctuple = matrix_constraint(A, x=vlist)
        self.assertEqual(len(ctuple.x), 5)
        self.assertEqual(len(ctuple.x), len(vlist))
        self.assertIsNot(ctuple.x, len(vlist))
        for i, v in enumerate(ctuple.x):
            self.assertIs(v, vlist[i])

        vlist = _create_variable_list(4)
        with self.assertRaises(ValueError):
            ctuple = matrix_constraint(A, x=vlist)
        ctuple = matrix_constraint(A)
        with self.assertRaises(ValueError):
            ctuple.x = vlist

    def test_bad_shape(self):
        A = numpy.array([[1, 2, 3], [1, 2, 3]])
        matrix_constraint(A)
        A = scipy.sparse.csr_matrix(numpy.array([[1, 2, 3], [1, 2, 3]]))
        matrix_constraint(A)
        A = numpy.array([1, 2, 3])
        with self.assertRaises(ValueError):
            matrix_constraint(A)
        A = numpy.ones((2, 2, 2))
        with self.assertRaises(ValueError):
            matrix_constraint(A)
        A = [1, 2, 3]
        with self.assertRaises(AttributeError):
            matrix_constraint(A)

    def test_equality(self):
        A = numpy.ones((5, 4))
        ctuple = matrix_constraint(A, rhs=1)
        self.assertTrue((ctuple.lb == 1).all())
        self.assertTrue((ctuple.ub == 1).all())
        self.assertTrue((ctuple.rhs == 1).all())
        self.assertTrue((ctuple.equality == True).all())
        for c in ctuple:
            self.assertEqual(c.lb, 1)
            self.assertEqual(c.ub, 1)
            self.assertEqual(c.equality, True)

        ctuple = matrix_constraint(A, rhs=numpy.zeros(5))
        self.assertTrue((ctuple.lb == 0).all())
        self.assertTrue((ctuple.ub == 0).all())
        self.assertTrue((ctuple.rhs == 0).all())
        self.assertTrue((ctuple.equality == True).all())
        for c in ctuple:
            self.assertEqual(c.lb, 0)
            self.assertEqual(c.ub, 0)
            self.assertEqual(c.equality, True)

        # can not set when equality is True
        with self.assertRaises(ValueError):
            ctuple.lb = 2
        with self.assertRaises(ValueError):
            ctuple.lb = ctuple.lb
        for c in ctuple:
            with self.assertRaises(ValueError):
                c.lb = 2
        # can not set when equality is True
        with self.assertRaises(ValueError):
            ctuple.ub = 2
        with self.assertRaises(ValueError):
            ctuple.ub = ctuple.ub
        for c in ctuple:
            with self.assertRaises(ValueError):
                c.ub = 2

        ctuple.equality = False
        # can not get when equality is False
        with self.assertRaises(ValueError):
            ctuple.rhs

        self.assertTrue((ctuple.lb == 0).all())
        self.assertTrue((ctuple.ub == 0).all())
        self.assertTrue((ctuple.equality == False).all())
        for c in ctuple:
            self.assertEqual(c.lb, 0)
            self.assertEqual(c.ub, 0)
            self.assertEqual(c.equality, False)

        # can not set to True, must set rhs to a value
        with self.assertRaises(ValueError):
            ctuple.equality = True
        for c in ctuple:
            with self.assertRaises(ValueError):
                c.equality = True

        ctuple.rhs = 3
        self.assertTrue((ctuple.lb == 3).all())
        self.assertTrue((ctuple.ub == 3).all())
        self.assertTrue((ctuple.equality == True).all())
        for c in ctuple:
            self.assertEqual(c.lb, 3)
            self.assertEqual(c.ub, 3)
            self.assertEqual(c.equality, True)

        with self.assertRaises(ValueError):
            ctuple.rhs = None
        for c in ctuple:
            with self.assertRaises(ValueError):
                ctuple.rhs = None
        self.assertTrue((ctuple.lb == 3).all())
        self.assertTrue((ctuple.ub == 3).all())
        self.assertTrue((ctuple.equality == True).all())
        for c in ctuple:
            self.assertEqual(c.lb, 3)
            self.assertEqual(c.ub, 3)
            self.assertEqual(c.equality, True)

        ctuple.equality = False
        self.assertTrue((ctuple.lb == 3).all())
        self.assertTrue((ctuple.ub == 3).all())
        self.assertTrue((ctuple.equality == False).all())
        for c in ctuple:
            self.assertEqual(c.lb, 3)
            self.assertEqual(c.ub, 3)
            self.assertEqual(c.equality, False)

        ctuple.rhs = 4
        self.assertTrue((ctuple.lb == 4).all())
        self.assertTrue((ctuple.ub == 4).all())
        self.assertTrue((ctuple.equality == True).all())
        for c in ctuple:
            self.assertEqual(c.lb, 4)
            self.assertEqual(c.ub, 4)
            self.assertEqual(c.equality, True)

        for c in ctuple:
            c.equality = False
        self.assertTrue((ctuple.lb == 4).all())
        self.assertTrue((ctuple.ub == 4).all())
        self.assertTrue((ctuple.equality == False).all())
        for c in ctuple:
            self.assertEqual(c.lb, 4)
            self.assertEqual(c.ub, 4)
            self.assertEqual(c.equality, False)

    def test_nondata_bounds(self):
        A = numpy.ones((5, 4))
        ctuple = matrix_constraint(A, rhs=1)

        eL = expression()
        eU = expression()
        with self.assertRaises(ValueError):
            ctuple.rhs = eL
        for c in ctuple:
            with self.assertRaises(ValueError):
                c.rhs = eL
        self.assertTrue((ctuple.rhs == 1).all())
        self.assertTrue((ctuple.equality == True).all())

        vL = variable()
        vU = variable()
        with self.assertRaises(ValueError):
            ctuple.rhs = vL
        for c in ctuple:
            with self.assertRaises(ValueError):
                c.rhs = vL
        self.assertTrue((ctuple.rhs == 1).all())
        self.assertTrue((ctuple.equality == True).all())

        vL.value = 1.0
        vU.value = 1.0
        with self.assertRaises(ValueError):
            ctuple.rhs = vL
        for c in ctuple:
            with self.assertRaises(ValueError):
                c.rhs = vL
        self.assertTrue((ctuple.rhs == 1).all())
        self.assertTrue((ctuple.equality == True).all())

        # the fixed status of a variable
        # does not change this restriction
        vL.fixed = True
        vU.fixed = True
        with self.assertRaises(ValueError):
            ctuple.rhs = vL
        for c in ctuple:
            with self.assertRaises(ValueError):
                c.rhs = vL
        self.assertTrue((ctuple.rhs == 1).all())
        self.assertTrue((ctuple.equality == True).all())

        p = parameter(value=0)
        with self.assertRaises(ValueError):
            ctuple.rhs = p
        for c in ctuple:
            with self.assertRaises(ValueError):
                c.rhs = p
        self.assertTrue((ctuple.rhs == 1).all())
        self.assertTrue((ctuple.equality == True).all())

        ctuple.equality = False
        self.assertTrue((ctuple.lb == 1).all())
        self.assertTrue((ctuple.ub == 1).all())
        self.assertTrue((ctuple.equality == False).all())

        eL = expression()
        eU = expression()
        with self.assertRaises(ValueError):
            ctuple.lb = eL
        with self.assertRaises(ValueError):
            ctuple.ub = eU
        for c in ctuple:
            with self.assertRaises(ValueError):
                c.lb = eL
            with self.assertRaises(ValueError):
                c.ub = eU
            with self.assertRaises(ValueError):
                c.bounds = (eL, eU)
        self.assertTrue((ctuple.lb == 1).all())
        self.assertTrue((ctuple.ub == 1).all())
        self.assertTrue((ctuple.equality == False).all())

        vL = variable()
        vU = variable()
        with self.assertRaises(ValueError):
            ctuple.lb = vL
        with self.assertRaises(ValueError):
            ctuple.ub = vU
        for c in ctuple:
            with self.assertRaises(ValueError):
                c.lb = vL
            with self.assertRaises(ValueError):
                c.ub = vU
            with self.assertRaises(ValueError):
                c.bounds = (vL, vU)
        self.assertTrue((ctuple.lb == 1).all())
        self.assertTrue((ctuple.ub == 1).all())
        self.assertTrue((ctuple.equality == False).all())

        vL.value = 1.0
        vU.value = 1.0
        with self.assertRaises(ValueError):
            ctuple.lb = vL
        with self.assertRaises(ValueError):
            ctuple.ub = vU
        for c in ctuple:
            with self.assertRaises(ValueError):
                c.lb = vL
            with self.assertRaises(ValueError):
                c.ub = vU
            with self.assertRaises(ValueError):
                c.bounds = (vL, vU)
        self.assertTrue((ctuple.lb == 1).all())
        self.assertTrue((ctuple.ub == 1).all())
        self.assertTrue((ctuple.equality == False).all())

        # the fixed status of a variable
        # does not change this restriction
        vL.fixed = True
        vU.fixed = True
        with self.assertRaises(ValueError):
            ctuple.lb = vL
        with self.assertRaises(ValueError):
            ctuple.ub = vU
        for c in ctuple:
            with self.assertRaises(ValueError):
                c.lb = vL
            with self.assertRaises(ValueError):
                c.ub = vU
            with self.assertRaises(ValueError):
                c.bounds = (vL, vU)
        self.assertTrue((ctuple.lb == 1).all())
        self.assertTrue((ctuple.ub == 1).all())
        self.assertTrue((ctuple.equality == False).all())

        p = parameter(value=0)
        with self.assertRaises(ValueError):
            ctuple.lb = p
        with self.assertRaises(ValueError):
            ctuple.ub = p
        for c in ctuple:
            with self.assertRaises(ValueError):
                c.lb = p
            with self.assertRaises(ValueError):
                c.ub = p
            with self.assertRaises(ValueError):
                c.bounds = (p, p)
        self.assertTrue((ctuple.lb == 1).all())
        self.assertTrue((ctuple.ub == 1).all())
        self.assertTrue((ctuple.equality == False).all())

    def test_data_bounds(self):
        A = numpy.ones((5, 4))
        ctuple = matrix_constraint(A)
        self.assertTrue((ctuple.lb == -numpy.inf).all())
        self.assertTrue((ctuple.ub == numpy.inf).all())
        self.assertTrue((ctuple.equality == False).all())
        with self.assertRaises(ValueError):
            ctuple.rhs
        for c in ctuple:
            self.assertEqual(c.lb, None)
            self.assertEqual(c.ub, None)
            self.assertEqual(c.equality, False)
            with self.assertRaises(ValueError):
                c.rhs

        ctuple.lb = 1
        ctuple.ub = 2
        self.assertTrue((ctuple.lb == 1).all())
        self.assertTrue((ctuple.ub == 2).all())
        self.assertTrue((ctuple.equality == False).all())
        with self.assertRaises(ValueError):
            ctuple.rhs
        for c in ctuple:
            self.assertEqual(c.lb, 1)
            self.assertEqual(c.ub, 2)
            self.assertEqual(c.equality, False)
            with self.assertRaises(ValueError):
                c.rhs

        with self.assertRaises(ValueError):
            ctuple.lb = range(5)
        with self.assertRaises(ValueError):
            ctuple.lb = numpy.array(range(6))
        with self.assertRaises(ValueError):
            ctuple.ub = range(5)
        with self.assertRaises(ValueError):
            ctuple.ub = numpy.array(range(6))
        with self.assertRaises(ValueError):
            ctuple.equality = True
        for c in ctuple:
            with self.assertRaises(ValueError):
                ctuple.equality = True

        self.assertTrue((ctuple.lb == 1).all())
        self.assertTrue((ctuple.ub == 2).all())
        self.assertTrue((ctuple.equality == False).all())
        with self.assertRaises(ValueError):
            ctuple.rhs
        for c in ctuple:
            self.assertEqual(c.lb, 1)
            self.assertEqual(c.ub, 2)
            self.assertEqual(c.equality, False)
            with self.assertRaises(ValueError):
                c.rhs

        for c in ctuple:
            c.bounds = (-1, 1)
        self.assertTrue((ctuple.lb == -1).all())
        self.assertTrue((ctuple.ub == 1).all())
        self.assertTrue((ctuple.equality == False).all())
        with self.assertRaises(ValueError):
            ctuple.rhs
        for c in ctuple:
            self.assertEqual(c.lb, -1)
            self.assertEqual(c.ub, 1)
            self.assertEqual(c.bounds, (-1, 1))
            self.assertEqual(c.equality, False)
            with self.assertRaises(ValueError):
                c.rhs

        ctuple.lb = lb_ = -numpy.array(range(5))
        ctuple.ub = ub_ = numpy.array(range(5))
        self.assertTrue((ctuple.lb == lb_).all())
        self.assertTrue((ctuple.ub == ub_).all())
        self.assertTrue((ctuple.equality == False).all())
        with self.assertRaises(ValueError):
            ctuple.rhs
        for i, c in enumerate(ctuple):
            self.assertEqual(c.lb, -i)
            self.assertEqual(c.ub, i)
            self.assertEqual(c.equality, False)
            with self.assertRaises(ValueError):
                c.rhs

        for c in ctuple:
            c.lb = None
            c.ub = None
        self.assertTrue((ctuple.lb == -numpy.inf).all())
        self.assertTrue((ctuple.ub == numpy.inf).all())
        self.assertTrue((ctuple.equality == False).all())
        with self.assertRaises(ValueError):
            ctuple.rhs
        for c in ctuple:
            self.assertEqual(c.lb, None)
            self.assertEqual(c.ub, None)
            self.assertEqual(c.equality, False)
            with self.assertRaises(ValueError):
                c.rhs

        for i, c in enumerate(ctuple):
            c.lb = -i
            c.ub = i
        self.assertTrue((ctuple.lb == lb_).all())
        self.assertTrue((ctuple.ub == ub_).all())
        self.assertTrue((ctuple.equality == False).all())
        with self.assertRaises(ValueError):
            ctuple.rhs
        for i, c in enumerate(ctuple):
            self.assertEqual(c.lb, -i)
            self.assertEqual(c.ub, i)
            self.assertEqual(c.equality, False)
            with self.assertRaises(ValueError):
                c.rhs

        ctuple.lb = None
        ctuple.ub = None
        self.assertTrue((ctuple.lb == -numpy.inf).all())
        self.assertTrue((ctuple.ub == numpy.inf).all())
        self.assertTrue((ctuple.equality == False).all())
        with self.assertRaises(ValueError):
            ctuple.rhs
        for c in ctuple:
            self.assertEqual(c.lb, None)
            self.assertEqual(c.ub, None)
            self.assertEqual(c.equality, False)
            with self.assertRaises(ValueError):
                c.rhs

        ctuple.rhs = 1
        self.assertTrue((ctuple.lb == 1).all())
        self.assertTrue((ctuple.ub == 1).all())
        self.assertTrue((ctuple.ub == 1).all())
        self.assertTrue((ctuple.equality == True).all())
        for c in ctuple:
            self.assertEqual(c.lb, 1)
            self.assertEqual(c.ub, 1)
            self.assertEqual(c.rhs, 1)
            self.assertEqual(c.equality, True)

        rhs_ = numpy.array(range(5))
        for i, c in enumerate(ctuple):
            c.rhs = i
        self.assertTrue((ctuple.lb == rhs_).all())
        self.assertTrue((ctuple.ub == rhs_).all())
        self.assertTrue((ctuple.ub == rhs_).all())
        self.assertTrue((ctuple.equality == True).all())
        for i, c in enumerate(ctuple):
            self.assertEqual(c.lb, i)
            self.assertEqual(c.ub, i)
            self.assertEqual(c.rhs, i)
            self.assertEqual(c.equality, True)

        with self.assertRaises(ValueError):
            ctuple.rhs = None
        with self.assertRaises(ValueError):
            ctuple.rhs = range(5)
        with self.assertRaises(ValueError):
            ctuple.rhs = numpy.array(range(6))
        for c in ctuple:
            with self.assertRaises(ValueError):
                c.rhs = None
        self.assertTrue((ctuple.lb == rhs_).all())
        self.assertTrue((ctuple.ub == rhs_).all())
        self.assertTrue((ctuple.ub == rhs_).all())
        self.assertTrue((ctuple.equality == True).all())
        for i, c in enumerate(ctuple):
            self.assertEqual(c.lb, i)
            self.assertEqual(c.ub, i)
            self.assertEqual(c.rhs, i)
            self.assertEqual(c.equality, True)

    def test_call(self):
        vlist = _create_variable_list(3)
        vlist[0].value = 1
        vlist[1].value = 0
        vlist[2].value = 3
        A = numpy.ones((3, 3))
        ctuple = matrix_constraint(A, x=vlist)
        self.assertTrue((ctuple() == 4).all())
        self.assertEqual(ctuple[0](), 4)
        self.assertEqual(ctuple[1](), 4)
        self.assertEqual(ctuple[2](), 4)
        A[:, 0] = 0
        A[:, 2] = 2
        ctuple = matrix_constraint(A, x=vlist)
        vlist[2].value = 4
        self.assertTrue((ctuple() == 8).all())
        self.assertEqual(ctuple[0](), 8)
        self.assertEqual(ctuple[1](), 8)
        self.assertEqual(ctuple[2](), 8)

        A = numpy.random.rand(4, 3)
        ctuple = matrix_constraint(A, x=vlist)
        vlist[1].value = 2
        cvals = numpy.array([ctuple[0](), ctuple[1](), ctuple[2](), ctuple[3]()])
        self.assertTrue((ctuple() == cvals).all())

        vlist[1].value = None
        with self.assertRaises(ValueError):
            ctuple()
        with self.assertRaises(ValueError):
            ctuple(exception=True)
        self.assertIs(ctuple(exception=False), None)
        for c in ctuple:
            with self.assertRaises(ValueError):
                c()
            with self.assertRaises(ValueError):
                c(exception=True)
            self.assertIs(c(exception=False), None)

        ctuple.x = None
        with self.assertRaises(ValueError):
            ctuple()
        with self.assertRaises(ValueError):
            ctuple(exception=True)
        with self.assertRaises(ValueError):
            ctuple(exception=False)
        for c in ctuple:
            with self.assertRaises(ValueError):
                c()
            with self.assertRaises(ValueError):
                c(exception=True)
            with self.assertRaises(ValueError):
                c(exception=False)

    def test_slack(self):
        vlist = _create_variable_list(3)
        vlist[0].value = 1
        vlist[1].value = 0
        vlist[2].value = 3
        A = numpy.ones((3, 3))
        ctuple = matrix_constraint(A, x=vlist)
        self.assertTrue((ctuple() == 4).all())
        self.assertEqual(ctuple[0](), 4)
        self.assertEqual(ctuple[1](), 4)
        self.assertEqual(ctuple[2](), 4)
        A[:, 0] = 0
        A[:, 2] = 2
        ctuple = matrix_constraint(A, x=vlist)
        vlist[2].value = 4
        self.assertTrue((ctuple() == 8).all())
        self.assertEqual(ctuple[0](), 8)
        self.assertEqual(ctuple[1](), 8)
        self.assertEqual(ctuple[2](), 8)

        A = numpy.random.rand(4, 3)
        ctuple = matrix_constraint(A, x=vlist)
        vlist[1].value = 2
        cvals = numpy.array([ctuple[0](), ctuple[1](), ctuple[2](), ctuple[3]()])
        self.assertTrue((ctuple() == cvals).all())

    def test_slack_methods(self):
        x = variable(value=2)
        L = 1
        U = 5
        A = numpy.array([[1]])

        cE = matrix_constraint(A, x=[x], rhs=L)
        x.value = 4
        self.assertEqual(cE[0].body(), 4)
        self.assertEqual(cE[0].slack, -3)
        self.assertEqual(cE[0].lslack, 3)
        self.assertEqual(cE[0].uslack, -3)
        self.assertEqual(cE.slack[0], -3)
        self.assertEqual(cE.lslack[0], 3)
        self.assertEqual(cE.uslack[0], -3)
        x.value = 6
        self.assertEqual(cE[0].body(), 6)
        self.assertEqual(cE[0].slack, -5)
        self.assertEqual(cE[0].lslack, 5)
        self.assertEqual(cE[0].uslack, -5)
        self.assertEqual(cE.slack[0], -5)
        self.assertEqual(cE.lslack[0], 5)
        self.assertEqual(cE.uslack[0], -5)
        x.value = 0
        self.assertEqual(cE[0].body(), 0)
        self.assertEqual(cE[0].slack, -1)
        self.assertEqual(cE[0].lslack, -1)
        self.assertEqual(cE[0].uslack, 1)
        self.assertEqual(cE.slack[0], -1)
        self.assertEqual(cE.lslack[0], -1)
        self.assertEqual(cE.uslack[0], 1)
        x.value = None
        with self.assertRaises(ValueError):
            cE[0].body()
        with self.assertRaises(ValueError):
            cE[0].body(exception=True)
        self.assertEqual(cE[0].body(exception=False), None)
        self.assertEqual(cE[0].slack, None)
        self.assertEqual(cE[0].lslack, None)
        self.assertEqual(cE[0].uslack, None)
        self.assertEqual(cE.slack, None)
        self.assertEqual(cE.lslack, None)
        self.assertEqual(cE.uslack, None)

        cE = matrix_constraint(A, x=[x], rhs=U)
        x.value = 4
        self.assertEqual(cE[0].body(), 4)
        self.assertEqual(cE[0].slack, -1)
        self.assertEqual(cE[0].lslack, -1)
        self.assertEqual(cE[0].uslack, 1)
        self.assertEqual(cE.slack[0], -1)
        self.assertEqual(cE.lslack[0], -1)
        self.assertEqual(cE.uslack[0], 1)
        x.value = 6
        self.assertEqual(cE[0].body(), 6)
        self.assertEqual(cE[0].slack, -1)
        self.assertEqual(cE[0].lslack, 1)
        self.assertEqual(cE[0].uslack, -1)
        self.assertEqual(cE.slack[0], -1)
        self.assertEqual(cE.lslack[0], 1)
        self.assertEqual(cE.uslack[0], -1)
        x.value = 0
        self.assertEqual(cE[0].body(), 0)
        self.assertEqual(cE[0].slack, -5)
        self.assertEqual(cE[0].lslack, -5)
        self.assertEqual(cE[0].uslack, 5)
        self.assertEqual(cE.slack[0], -5)
        self.assertEqual(cE.lslack[0], -5)
        self.assertEqual(cE.uslack[0], 5)

        cL = matrix_constraint(A, x=[x], lb=L)
        x.value = 4
        self.assertEqual(cL[0].body(), 4)
        self.assertEqual(cL[0].slack, 3)
        self.assertEqual(cL[0].lslack, 3)
        self.assertEqual(cL[0].uslack, float('inf'))
        self.assertEqual(cL.slack[0], 3)
        self.assertEqual(cL.lslack[0], 3)
        self.assertEqual(cL.uslack[0], float('inf'))
        x.value = 6
        self.assertEqual(cL[0].body(), 6)
        self.assertEqual(cL[0].slack, 5)
        self.assertEqual(cL[0].lslack, 5)
        self.assertEqual(cL[0].uslack, float('inf'))
        self.assertEqual(cL.slack[0], 5)
        self.assertEqual(cL.lslack[0], 5)
        self.assertEqual(cL.uslack[0], float('inf'))
        x.value = 0
        self.assertEqual(cL[0].body(), 0)
        self.assertEqual(cL[0].slack, -1)
        self.assertEqual(cL[0].lslack, -1)
        self.assertEqual(cL[0].uslack, float('inf'))
        self.assertEqual(cL.slack[0], -1)
        self.assertEqual(cL.lslack[0], -1)
        self.assertEqual(cL.uslack[0], float('inf'))

        cL = matrix_constraint(A, x=[x], lb=float('-inf'))
        x.value = 4
        self.assertEqual(cL[0].body(), 4)
        self.assertEqual(cL[0].slack, float('inf'))
        self.assertEqual(cL[0].lslack, float('inf'))
        self.assertEqual(cL[0].uslack, float('inf'))
        self.assertEqual(cL.slack[0], float('inf'))
        self.assertEqual(cL.lslack[0], float('inf'))
        self.assertEqual(cL.uslack[0], float('inf'))
        x.value = 6
        self.assertEqual(cL[0].body(), 6)
        self.assertEqual(cL[0].slack, float('inf'))
        self.assertEqual(cL[0].lslack, float('inf'))
        self.assertEqual(cL[0].uslack, float('inf'))
        self.assertEqual(cL.slack[0], float('inf'))
        self.assertEqual(cL.lslack[0], float('inf'))
        self.assertEqual(cL.uslack[0], float('inf'))
        x.value = 0
        self.assertEqual(cL[0].body(), 0)
        self.assertEqual(cL[0].slack, float('inf'))
        self.assertEqual(cL[0].lslack, float('inf'))
        self.assertEqual(cL[0].uslack, float('inf'))
        self.assertEqual(cL.slack[0], float('inf'))
        self.assertEqual(cL.lslack[0], float('inf'))
        self.assertEqual(cL.uslack[0], float('inf'))

        cU = matrix_constraint(A, x=[x], ub=U)
        x.value = 4
        self.assertEqual(cU[0].body(), 4)
        self.assertEqual(cU[0].slack, 1)
        self.assertEqual(cU[0].lslack, float('inf'))
        self.assertEqual(cU[0].uslack, 1)
        self.assertEqual(cU.slack[0], 1)
        self.assertEqual(cU.lslack[0], float('inf'))
        self.assertEqual(cU.uslack[0], 1)
        x.value = 6
        self.assertEqual(cU[0].body(), 6)
        self.assertEqual(cU[0].slack, -1)
        self.assertEqual(cU[0].lslack, float('inf'))
        self.assertEqual(cU[0].uslack, -1)
        self.assertEqual(cU.slack[0], -1)
        self.assertEqual(cU.lslack[0], float('inf'))
        self.assertEqual(cU.uslack[0], -1)
        x.value = 0
        self.assertEqual(cU[0].body(), 0)
        self.assertEqual(cU[0].slack, 5)
        self.assertEqual(cU[0].lslack, float('inf'))
        self.assertEqual(cU[0].uslack, 5)
        self.assertEqual(cU.slack[0], 5)
        self.assertEqual(cU.lslack[0], float('inf'))
        self.assertEqual(cU.uslack[0], 5)

        cU = matrix_constraint(A, x=[x], ub=float('inf'))
        x.value = 4
        self.assertEqual(cU[0].body(), 4)
        self.assertEqual(cU[0].slack, float('inf'))
        self.assertEqual(cU[0].lslack, float('inf'))
        self.assertEqual(cU[0].uslack, float('inf'))
        self.assertEqual(cU.slack[0], float('inf'))
        self.assertEqual(cU.lslack[0], float('inf'))
        self.assertEqual(cU.uslack[0], float('inf'))
        x.value = 6
        self.assertEqual(cU[0].body(), 6)
        self.assertEqual(cU[0].slack, float('inf'))
        self.assertEqual(cU[0].lslack, float('inf'))
        self.assertEqual(cU[0].uslack, float('inf'))
        self.assertEqual(cU.slack[0], float('inf'))
        self.assertEqual(cU.lslack[0], float('inf'))
        self.assertEqual(cU.uslack[0], float('inf'))
        x.value = 0
        self.assertEqual(cU[0].body(), 0)
        self.assertEqual(cU[0].slack, float('inf'))
        self.assertEqual(cU[0].lslack, float('inf'))
        self.assertEqual(cU[0].uslack, float('inf'))
        self.assertEqual(cU.slack[0], float('inf'))
        self.assertEqual(cU.lslack[0], float('inf'))
        self.assertEqual(cU.uslack[0], float('inf'))

        cR = matrix_constraint(A, x=[x], lb=L, ub=U)
        x.value = 4
        self.assertEqual(cR[0].body(), 4)
        self.assertEqual(cR[0].slack, 1)
        self.assertEqual(cR[0].lslack, 3)
        self.assertEqual(cR[0].uslack, 1)
        self.assertEqual(cR.slack[0], 1)
        self.assertEqual(cR.lslack[0], 3)
        self.assertEqual(cR.uslack[0], 1)
        x.value = 6
        self.assertEqual(cR[0].body(), 6)
        self.assertEqual(cR[0].slack, -1)
        self.assertEqual(cR[0].lslack, 5)
        self.assertEqual(cR[0].uslack, -1)
        self.assertEqual(cR.slack[0], -1)
        self.assertEqual(cR.lslack[0], 5)
        self.assertEqual(cR.uslack[0], -1)
        x.value = 0
        self.assertEqual(cR[0].body(), 0)
        self.assertEqual(cR[0].slack, -1)
        self.assertEqual(cR[0].lslack, -1)
        self.assertEqual(cR[0].uslack, 5)
        self.assertEqual(cR.slack[0], -1)
        self.assertEqual(cR.lslack[0], -1)
        self.assertEqual(cR.uslack[0], 5)

        cR = matrix_constraint(A, x=[x])
        x.value = 4
        self.assertEqual(cR[0].body(), 4)
        self.assertEqual(cR[0].slack, float('inf'))
        self.assertEqual(cR[0].lslack, float('inf'))
        self.assertEqual(cR[0].uslack, float('inf'))
        self.assertEqual(cR.slack[0], float('inf'))
        self.assertEqual(cR.lslack[0], float('inf'))
        self.assertEqual(cR.uslack[0], float('inf'))
        x.value = 6
        self.assertEqual(cR[0].body(), 6)
        self.assertEqual(cR[0].slack, float('inf'))
        self.assertEqual(cR[0].lslack, float('inf'))
        self.assertEqual(cR[0].uslack, float('inf'))
        self.assertEqual(cR.slack[0], float('inf'))
        self.assertEqual(cR.lslack[0], float('inf'))
        self.assertEqual(cR.uslack[0], float('inf'))
        x.value = 0
        self.assertEqual(cR[0].body(), 0)
        self.assertEqual(cR[0].slack, float('inf'))
        self.assertEqual(cR[0].lslack, float('inf'))
        self.assertEqual(cR[0].uslack, float('inf'))
        self.assertEqual(cR.slack[0], float('inf'))
        self.assertEqual(cR.lslack[0], float('inf'))
        self.assertEqual(cR.uslack[0], float('inf'))

        cR = matrix_constraint(A, x=[x], lb=float('-inf'), ub=float('inf'))
        x.value = 4
        self.assertEqual(cR[0].body(), 4)
        self.assertEqual(cR[0].slack, float('inf'))
        self.assertEqual(cR[0].lslack, float('inf'))
        self.assertEqual(cR[0].uslack, float('inf'))
        self.assertEqual(cR.slack[0], float('inf'))
        self.assertEqual(cR.lslack[0], float('inf'))
        self.assertEqual(cR.uslack[0], float('inf'))
        x.value = 6
        self.assertEqual(cR[0].body(), 6)
        self.assertEqual(cR[0].slack, float('inf'))
        self.assertEqual(cR[0].lslack, float('inf'))
        self.assertEqual(cR[0].uslack, float('inf'))
        self.assertEqual(cR.slack[0], float('inf'))
        self.assertEqual(cR.lslack[0], float('inf'))
        self.assertEqual(cR.uslack[0], float('inf'))
        x.value = 0
        self.assertEqual(cR[0].body(), 0)
        self.assertEqual(cR[0].slack, float('inf'))
        self.assertEqual(cR[0].lslack, float('inf'))
        self.assertEqual(cR[0].uslack, float('inf'))
        self.assertEqual(cR.slack[0], float('inf'))
        self.assertEqual(cR.lslack[0], float('inf'))
        self.assertEqual(cR.uslack[0], float('inf'))

    def test_canonical_form_sparse(self):
        A = numpy.array([[0, 2]])
        vlist = _create_variable_list(2)
        ctuple = matrix_constraint(A, x=vlist)
        self.assertEqual(ctuple.sparse, True)
        for c in ctuple:
            self.assertEqual(c._linear_canonical_form, True)
        terms = list(ctuple[0].terms)
        vs, cs = zip(*terms)
        self.assertEqual(len(terms), 1)
        self.assertIs(vs[0], vlist[1])
        self.assertEqual(cs[0], 2)
        repn = ctuple[0].canonical_form()
        self.assertEqual(len(repn.linear_vars), 1)
        self.assertIs(repn.linear_vars[0], vlist[1])
        self.assertEqual(repn.linear_coefs, (2,))
        self.assertEqual(repn.constant, 0)
        vlist[0].fix(1)
        repn = ctuple[0].canonical_form()
        self.assertEqual(len(repn.linear_vars), 1)
        self.assertIs(repn.linear_vars[0], vlist[1])
        self.assertEqual(repn.linear_coefs, (2,))
        self.assertEqual(repn.constant, 0)
        vlist[1].fix(2)
        repn = ctuple[0].canonical_form()
        self.assertEqual(repn.linear_vars, ())
        self.assertEqual(repn.linear_coefs, ())
        self.assertEqual(repn.constant, 4)
        repn = ctuple[0].canonical_form(compute_values=False)
        self.assertEqual(repn.linear_vars, ())
        self.assertEqual(repn.linear_coefs, ())
        self.assertEqual(repn.constant(), 4)

    def test_canonical_form_dense(self):
        A = numpy.array([[0, 2]])
        vlist = _create_variable_list(2)
        ctuple = matrix_constraint(A, x=vlist, sparse=False)
        self.assertEqual(ctuple.sparse, False)
        for c in ctuple:
            self.assertEqual(c._linear_canonical_form, True)
        terms = list(ctuple[0].terms)
        vs, cs = zip(*terms)
        self.assertEqual(len(terms), 2)
        self.assertIs(vs[0], vlist[0])
        self.assertIs(vs[1], vlist[1])
        self.assertEqual(cs[0], 0)
        self.assertEqual(cs[1], 2)
        repn = ctuple[0].canonical_form()
        self.assertEqual(len(repn.linear_vars), 2)
        self.assertIs(repn.linear_vars[0], vlist[0])
        self.assertIs(repn.linear_vars[1], vlist[1])
        self.assertEqual(repn.linear_coefs, (0, 2))
        self.assertEqual(repn.constant, 0)
        vlist[0].fix(1)
        repn = ctuple[0].canonical_form()
        self.assertEqual(len(repn.linear_vars), 1)
        self.assertIs(repn.linear_vars[0], vlist[1])
        self.assertEqual(repn.linear_coefs, (2,))
        self.assertEqual(repn.constant, 0)
        vlist[1].fix(2)
        repn = ctuple[0].canonical_form()
        self.assertEqual(repn.linear_vars, ())
        self.assertEqual(repn.linear_coefs, ())
        self.assertEqual(repn.constant, 4)
        repn = ctuple[0].canonical_form(compute_values=False)
        self.assertEqual(repn.linear_vars, ())
        self.assertEqual(repn.linear_coefs, ())
        self.assertEqual(repn.constant(), 4)

    def test_preorder_traversal(self):
        A = numpy.ones((3, 3))

        m = block()
        m.c = matrix_constraint(A)
        m.v = variable()
        m.V = variable_list()
        m.V.append(variable())
        m.B = block_list()
        m.B.append(block())
        m.B[0].c = matrix_constraint(A)
        m.B[0].v = variable()
        m.B[0].V = variable_list()
        m.B[0].V.append(variable())
        m.b = block()
        m.b.c = constraint_dict()
        m.b.c[None] = matrix_constraint(A)
        m.b.c[1] = matrix_constraint(A)
        m.b.c[2] = constraint()
        m.b.c[3] = constraint_list()

        # don't visit things below a matrix constraint
        # (e.g., cases where we want to handle it in bulk)
        def no_mc_descend(x):
            if isinstance(x, matrix_constraint):
                return False
            return True

        cnt = 0
        for obj in pmo.preorder_traversal(m, ctype=IConstraint, descend=no_mc_descend):
            self.assertTrue(type(obj.parent) is not matrix_constraint)
            self.assertTrue(
                (obj.ctype is block._ctype) or (obj.ctype is constraint._ctype)
            )
            cnt += 1
        self.assertEqual(cnt, 11)

        cnt = 0
        mc_child_cnt = 0
        for obj in pmo.preorder_traversal(m, ctype=IConstraint):
            self.assertTrue(
                (obj.ctype is block._ctype) or (obj.ctype is constraint._ctype)
            )
            if type(obj.parent) is matrix_constraint:
                mc_child_cnt += 1
            cnt += 1
        self.assertEqual(cnt, 23)
        self.assertEqual(mc_child_cnt, 12)

        self.assertEqual(len(list(m.components(ctype=IConstraint))), 13)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os.path
import re
import sys
import shutil
from decimal import Decimal

from pyomo.common.dependencies import attempt_import
from pyomo.dataportal import TableData
from pyomo.dataportal.factory import DataManagerFactory

pyodbc, pyodbc_available = attempt_import('pyodbc')
pypyodbc, pypyodbc_available = attempt_import('pypyodbc')
sqlite3, sqlite3_available = attempt_import('sqlite3')
pymysql, pymysql_available = attempt_import('pymysql')

# format=
# using=
# query=
# user=
# password=
# table=


class db_Table(TableData):
    def __init__(self):
        TableData.__init__(self)
        self.using = None

    def open(self):
        if self.filename is None:
            raise IOError("No data source name was specified")
        if self.filename[0] == '"':
            self.filename = self.filename[1:-1]
        #
        # Initialize self.db
        #
        if not self.using is None:
            self.options.using = self.using
        self.db = None
        if self._data is not None:
            self.db = self._data
        else:
            try:
                self.db = self.connect(self.filename, self.options)
            except Exception:
                raise

    def read(self):
        #
        # Get the table from the database
        #
        if self.db is None:
            return
        cursor = self.db.cursor()
        tmp = []
        if self.options.query is None:
            if self.options.table is None:
                raise IOError("Must specify 'query' or 'table' option!")
            self.options.query = 'SELECT * FROM %s' % self.options.table
        elif self.options.query[0] in ("'", '"'):
            self.options.query = self.options.query[1:-1]

        try:
            cursor.execute(self.options.query)
            rows = cursor.fetchall()
            for col in cursor.description:
                tmp.append(col[0])
            tmp = [tmp]
        except sqlite3.OperationalError:
            import logging

            logging.getLogger('pyomo.core').error(
                """Fatal error reading from an external ODBC data source.

This error was generated outside Pyomo by the Python connector to the
external data source:

    %s

for the query:

    %s

It is possible that you have an error in your external data file,
the ODBC connector for this data source is not correctly installed,
or that there is a bug in the ODBC connector.
""" % (self.filename, self.options.query)
            )
            raise
        for row in rows:
            # print("DATA %s" % str(list(row))) # XXX
            ttmp = []
            for data in list(row):
                if isinstance(data, Decimal):
                    ttmp.append(float(data))
                elif data is None:
                    ttmp.append('.')
                elif isinstance(data, str):
                    nulidx = data.find('\x00')
                    if nulidx > -1:
                        data = data[:nulidx]
                    ttmp.append(data)
                else:
                    ttmp.append(data)
            tmp.append(ttmp)
        # print('FINAL %s' % str(tmp)) # XXX
        #
        # Process data from the table
        #
        if type(tmp) in (int, float):
            if not self.options.param is None:
                self._info = ["param", self.options.param.local_name, ":=", tmp]
            elif len(self.options.symbol_map) == 1:
                self._info = [
                    "param",
                    self.options.symbol_map[self.options.symbol_map.keys()[0]],
                    ":=",
                    tmp,
                ]
            else:
                raise IOError(
                    "Data looks like a scalar parameter, but multiple parameter names have been specified: %s"
                    % str(self.options.symbol_map)
                )
        elif len(tmp) == 0:
            raise IOError("Empty range '%s'" % self.options.range)
        else:
            # print("_info %s" % str(self._info))
            # print("SETTING DATA %s %s" % (str(tmp[0]), str(tmp[1:]))) # XXX
            # print("OPTIONS %s" % str(self.options))
            self._set_data(tmp[0], tmp[1:])

    def close(self):
        if self._data is None and not self.db is None:
            del self.db

    def connect(self, connection, options, kwds={}):
        try:
            mod = __import__(options.using)
            args = [connection]
            if not options.user is None:
                args.append(options.user)
            if not options.password is None:
                args.append(options.password)
            if not options.database is None:
                args.append(options.database)
            return mod.connect(*args, **kwds)
        except ImportError:
            return None


#
# NOTE: The pyodbc interface currently doesn't work.  Notably, nothing
# sets the "table" or "query" options, which causes db_table.read() to
# fail.  This interface has been disabled by overriding pyodbc_available
# in sheet.py
#


@DataManagerFactory.register('pyodbc', "%s database interface" % 'pyodbc')
class pyodbc_db_Table(db_Table):
    _drivers = {
        'mdb': ["Microsoft Access Driver (*.mdb)"],
        'xls': [
            "Microsoft Excel Driver (*.xls, *.xlsx, *.xlsm, *.xlsb)",
            "Microsoft Excel Driver (*.xls)",
        ],
        'xlsx': ["Microsoft Excel Driver (*.xls, *.xlsx, *.xlsm, *.xlsb)"],
        'xlsm': ["Microsoft Excel Driver (*.xls, *.xlsx, *.xlsm, *.xlsb)"],
        'xlsb': ["Microsoft Excel Driver (*.xls, *.xlsx, *.xlsm, *.xlsb)"],
        'mysql': ["MySQL"],
    }
    _drivers['access'] = _drivers['mdb']
    _drivers['excel'] = _drivers['xls']

    def __init__(self):
        db_Table.__init__(self)
        self.using = 'pyodbc'

    def available(self):
        return pyodbc_available

    def requirements(self):
        return 'pyodbc'

    def connect(self, connection, options):
        if not options.driver is None:
            ctype = options.driver
        elif '.' in connection:
            ctype = connection.split('.')[-1]
        elif 'mysql' in connection.lower():
            ctype = 'mysql'
        else:
            ctype = ''
        extras = {}
        if (
            ctype in ['xls', 'xlsx', 'xlsm', 'xlsb', 'excel']
            or '.xls' in connection
            or '.xlsx' in connection
            or '.xlsm' in connection
            or '.xlsb' in connection
        ):
            extras['autocommit'] = True

        connection = self.create_connection_string(ctype, connection, options)
        try:
            conn = db_Table.connect(self, connection, options, extras)
        except TypeError:
            raise
        except Exception:
            e = sys.exc_info()[1]
            code = e.args[0]
            if code == 'IM002' or code == '08001':
                if 'HOME' in os.environ:
                    # Need a DSN! Try to add it to $HOME/.odbc.ini ...
                    #
                    # Note: this only works on *nix platforms.  It appears
                    # that ODBC.INI is stored in the registry on windows
                    #
                    # [JDS, 8 Oct 18]: I am not convinced that writing a
                    # .odbc.ini file is necessary; see the "else" branch
                    # below.
                    #
                    odbcIniPath = os.path.join(os.environ['HOME'], '.odbc.ini')
                    if os.path.exists(odbcIniPath):
                        shutil.copy(odbcIniPath, odbcIniPath + '.orig')
                        config = ODBCConfig(filename=odbcIniPath)
                    else:
                        config = ODBCConfig()

                    dsninfo = self.create_dsn_dict(connection, config)
                    dsnid = re.sub(
                        '[^A-Za-z0-9]', '', dsninfo['Database']
                    )  # Strip filenames of funny characters
                    dsn = 'PYOMO{0}'.format(dsnid)

                    config.add_source(dsn, dsninfo['Driver'])
                    config.add_source_spec(dsn, dsninfo)
                    config.write(odbcIniPath)

                    connstr = "DRIVER={{{0}}};DSN={1}".format(dsninfo['Driver'], dsn)
                else:
                    # Attempt to re-generate the connection string with a Driver
                    config = ODBCConfig()
                    dsninfo = self.create_dsn_dict(connection, config)
                    connstr = []
                    for k, v in dsninfo.items():
                        if ' ' in v and (v[0] != "{" or v[-1] != "}"):
                            connstr.append("%s={%s}" % (k.upper(), v))
                        else:
                            connstr.append("%s=%s" % (k.upper(), v))
                    connstr = ";".join(connstr)

                conn = db_Table.connect(
                    self, connstr, options, extras
                )  # Will raise its own exception on failure

            # Propagate the exception
            else:
                raise

        return conn

    def create_dsn_dict(self, argstr, existing_config):
        result = {}

        parts = argstr.split(';')
        argdict = {}
        for part in parts:
            if len(part) > 0 and '=' in part:
                key, val = part.split('=', 1)
                argdict[key.lower().strip()] = val.strip()

        if 'driver' in argdict:
            result['Driver'] = "{0}".format(argdict['driver']).strip("{}")

        if 'dsn' in argdict:
            if argdict['dsn'] in existing_config.source_specs:
                return existing_config.source_specs[argdict['dsn']]
            else:
                import logging

                logger = logging.getLogger("pyomo.core")
                logger.warning(
                    "DSN with name {0} not found. Attempting to continue with options...".format(
                        argdict['dsn']
                    )
                )

        if 'dbq' in argdict:
            # Using a file for db access.
            if 'Driver' not in result:
                result['Driver'] = self._drivers[argdict['dbq'].split('.')[-1].lower()]
            result['Database'] = argdict['dbq']
            result['Server'] = 'localhost'
            result['User'] = ''
            result['Password'] = ''
            result['Port'] = '5432'
            result['Description'] = argdict['dbq']
            for k in argdict.keys():
                if k.capitalize() not in result:
                    result[k.capitalize()] = argdict[k]
        else:
            if 'Driver' not in result:
                raise Exception("No driver specified, and no DBQ to infer from")
            elif result['Driver'].lower() == "mysql":
                result['Driver'] = "MySQL"
                result['Server'] = argdict.get('server', 'localhost')
                result['Database'] = argdict.get('database', '')
                result['Port'] = argdict.get('port', '3306')
                result['Socket'] = argdict.get('socket', '')
                result['Option'] = argdict.get('option', '')
                result['Stmt'] = argdict.get('stmt', '')
                result['User'] = argdict.get('user', '')
                result['Password'] = argdict.get('password', '')
                result['Description'] = argdict.get('description', '')
            else:
                raise Exception(
                    "Unknown driver type '{0}' for database connection".format(
                        result['Driver']
                    )
                )

        return result

    def create_connection_string(self, ctype, connection, options):
        driver = self._get_driver(ctype)
        if driver:
            if ' ' in driver and (driver[0] != "{" or driver[-1] != "}"):
                return "DRIVER={%s};Dbq=%s;" % (driver, connection)
            else:
                return "DRIVER=%s;Dbq=%s;" % (driver, connection)
        return connection

    def _get_driver(self, ctype):
        # Given a list of possible drivers for this ctype, look to find
        # a match in the pyodbc.drivvers() list.  If a match is found,
        # return it.  Otherwise (arbitrarily) return the first one.  If
        # the ctype is not known, return None.
        drivers = self._drivers.get(ctype, [])
        for driver in drivers:
            if driver in pyodbc.drivers():
                return driver
        if drivers:
            return drivers[0]
        else:
            return None


class ODBCError(Exception):
    def __init__(self, value):
        self.parameter = value

    def __repr__(self):
        return repr(self.parameter)


class ODBCConfig:
    """
    Encapsulates an ODBC configuration file, usually odbc.ini or
    .odbc.ini, as specified by IBM. ODBC config data can be loaded
    either from a file or a string containing the relevant formatted
    data. Calling load() after initialization will update existing
    information in the config object with the new information.
    """

    def __init__(self, filename=None, data=None):
        """
        Create a new ODBC config instance, loading data from
        the given file and/or data string. Once initialized, the
        new config will contain the data represented in both
        arguments, if any. Data specified as a string argument
        will override that in the file.
        """

        # Hardcoded string required here.
        # See documentation:
        # https://www.ibm.com/docs/en/informix-servers/12.10?topic=SSGU8G_12.1.0/com.ibm.odbc.doc/ids_odbc_062.html
        self.ODBC_DS_KEY = 'ODBC Data Sources'
        self.ODBC_INFO_KEY = 'ODBC'

        self.file = filename

        self.sources = {}
        self.source_specs = {}
        self.odbc_info = {}

        self.load(self.file, data)

    def load(self, filename=None, data=None):
        """
        Load data from the given file and/or data string. If
        both are given, data contained in the string will override
        that in the file. If this config object already contains
        data, the new information loaded will update the old,
        replacing where keys are the same.
        """

        sections = {}
        if filename is not None:
            with open(filename, 'r') as fileHandle:
                fileData = fileHandle.read()
                sections.update(self._get_sections(fileData))
        if data is not None:
            sections.update(self._get_sections(data))

        if self.ODBC_DS_KEY in sections:
            self.sources.update(sections[self.ODBC_DS_KEY])
            del sections[self.ODBC_DS_KEY]

        if self.ODBC_INFO_KEY in sections:
            self.odbc_info.update(sections[self.ODBC_INFO_KEY])
            del sections[self.ODBC_INFO_KEY]

        self.source_specs.update(sections)

    def __str__(self):
        return "<ODBC config: {0} sources, {1} source specs>".format(
            len(self.sources), len(self.source_specs)
        )

    def __eq__(self, other):
        if isinstance(other, ODBCConfig):
            return (
                self.sources == other.sources
                and self.source_specs == other.source_specs
                and self.odbc_info == other.odbc_info
            )
        return False

    def odbc_repr(self):
        """
        Get the full, odbc.ini-style representation of this
        ODBC configuration.
        """

        str = "[{0}]\n".format(self.ODBC_DS_KEY)

        for name in self.sources:
            str += "{0} = {1}\n".format(name, self.sources[name])

        for name in self.source_specs:
            str += "\n[{0}]\n".format(name)
            for key in self.source_specs[name]:
                str += "{0} = {1}\n".format(key, self.source_specs[name][key])

        if len(self.odbc_info) > 0:
            str += "\n[{0}]\n".format(self.ODBC_INFO_KEY)
            for key in self.odbc_info:
                str += "{0} = {1}\n".format(key, self.odbc_info[key])

        return str

    def write(self, filename):
        """
        Write the current ODBC configuration to the given file.
        Depends on the odbc_repr() function for a string
        representation of the stored ODBC config information.
        """
        with open(filename, 'w') as f:
            f.write(self.odbc_repr())

    def add_source(self, name, driver):
        """
        Add an ODBC data source to the configuration. A data
        source consists of a unique source name and a driver
        string, which specifies how the source will be loaded.
        If a source name is not unique, it will replace the
        existing source of the same name. A source is required
        in order to have a source specification.
        """

        if name is None or driver is None or len(name) == 0 or len(driver) == 0:
            raise ODBCError("A source must specify both a name and a driver string")
        if name == self.ODBC_DS_KEY or name == self.ODBC_INFO_KEY:
            raise ODBCError("A source cannot use the reserved name '{0}'".format(name))
        self.sources[str(name)] = str(driver)

    def del_source(self, name):
        """
        Remove an ODBC data source from the configuration. If
        any source specifications are based on this source, they
        will be removed as well.
        """

        if name in self.sources:
            if name in self.source_specs:
                del self.source_specs[name]
            del self.sources[name]

    def add_source_spec(self, name, spec):
        """
        Add an ODBC data source specification to the configuration.
        A source specification consists of a unique name and
        a key-value mapping (i.e. dictionary) of options. In order
        to add a source specification, a data source with a matching
        name must exist in the configuration.
        """

        if name is None or spec is None or len(name) == 0:
            raise ODBCError(
                "A source spec must specify both a name and a spec dictionary"
            )
        if name not in self.sources:
            raise ODBCError(
                "A source spec must have a corresponding source; call .add_source() first"
            )

        self.source_specs[name] = dict(spec)

    def del_source_spec(self, name):
        """
        Remove an ODBC data source specification from the
        configuration.
        """

        if name in self.source_specs:
            del self.source_specs[name]

    def set_odbc_info(self, key, value):
        """
        Set an option for the ODBC handling specified in the
        configuration. An option consists of a key-value pair.
        Specifying an existing key will update the current value.
        """

        if key is None or value is None or len(key) == 0 or len(value) == 0:
            raise ODBCError("An ODBC info pair must specify both a key and a value")

        self.odbc_info[str(key)] = str(value)

    def _get_sections(self, data):
        """
        Parse a string for ODBC sections. The parsing algorithm proceeds
        roughly as follows:

        1. Split the string on newline ('\\n') characters.
        2. Remove lines consisting purely of whitespace.
        3. Iterate over lines, storing all key-value pair lines in a dictionary.
        4. When reaching a new section header (denoted by '[str]'), store the old
           key-value pairs under the old section name. Continue from step 3.
        5. On reaching end of data, store the last section and return a mapping
           from section names to dictionaries of key-value pairs in those sections.
        """

        sections = {}
        sectionKey = None
        sectionContents = {}

        emptyLine = re.compile('^[ \t\r]*$')

        lines = data.split('\n')
        for line in lines:
            if emptyLine.match(line):
                # Whitespace only
                pass
            elif len(line) < 2:
                # Not enough room for even 'k='; can't contain info
                raise ODBCError(
                    "Malformed line in ODBC config (no meaningful data): " + line
                )
            elif line[0] == '[' and line[-1] == ']':
                # Starts a new section; '=' has no special meaning here
                sections[sectionKey] = sectionContents
                sectionKey = line[1:-1]
                sectionContents = {}
            else:
                # Not whitespace or section header; must be key=value. No duplicate '=' permitted.
                if '=' not in line:
                    raise ODBCError(
                        "Malformed line in ODBC config (no key-value mapping): " + line
                    )

                key, value = line.split("=", 1)
                if '=' in value:
                    raise ODBCError(
                        "Malformed line in ODBC config (too many '='): " + line
                    )

                sectionContents[key.strip()] = value.strip()
        sections[sectionKey] = sectionContents

        if None in sections:
            del sections[None]

        return sections


@DataManagerFactory.register('pypyodbc', "%s database interface" % 'pypyodbc')
class pypyodbc_db_Table(pyodbc_db_Table):
    def __init__(self):
        pyodbc_db_Table.__init__(self)
        self.using = 'pypyodbc'

    def available(self):
        return pypyodbc_available

    def requirements(self):
        return 'pypyodbc'

    def connect(self, connection, options):
        assert options['using'] == 'pypyodbc'

        return pyodbc_db_Table.connect(self, connection, options)


@DataManagerFactory.register('sqlite3', "sqlite3 database interface")
class sqlite3_db_Table(db_Table):
    def __init__(self):
        db_Table.__init__(self)
        self.using = 'sqlite3'

    def available(self):
        return sqlite3_available

    def requirements(self):
        return 'sqlite3'

    def connect(self, connection, options):
        assert options['using'] == 'sqlite3'

        filename = connection
        if not os.path.exists(filename):
            raise Exception("No such file: " + filename)

        con = sqlite3.connect(filename)
        if options.text_factory:
            con.text_factory = options.text_factory
        return con


@DataManagerFactory.register('pymysql', "pymysql database interface")
class pymysql_db_Table(db_Table):
    def __init__(self):
        db_Table.__init__(self)
        self.using = 'pymysql'

    def available(self):
        return pymysql_available

    def requirements(self):
        return 'pymysql'



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from collections import defaultdict
from threading import get_ident, main_thread


class MultiThreadWrapper:
    """A python object proxy that wraps different instances for each thread.

    This is useful for handling thread-safe access to singleton objects without
    having to refactor the code that depends on them.

    Note that instances of the wrapped object are reused if two threads share the same
    identifier, because identifiers could be reused and are unique only for any given moment.
    See [get_ident()](https://docs.python.org/3/library/threading.html#threading.get_ident) for more information.
    """

    __slots__ = '_mtdict'

    def __init__(self, base):
        object.__setattr__(self, '_mtdict', defaultdict(base))

    def __getattr__(self, attr):
        return getattr(self._mtdict[get_ident()], attr)

    def __setattr__(self, attr, value):
        setattr(self._mtdict[get_ident()], attr, value)

    def __delattr__(self, attr):
        delattr(self._mtdict[get_ident()], attr)

    def __enter__(self):
        return self._mtdict[get_ident()].__enter__()

    def __exit__(self, exc_type, exc_value, traceback):
        return self._mtdict[get_ident()].__exit__(exc_type, exc_value, traceback)

    def __dir__(self):
        return list(object.__dir__(self)) + list(self._mtdict[get_ident()].__dir__())

    def __str__(self):
        return self._mtdict[get_ident()].__str__()

    def __new__(cls, wrapped):
        return super().__new__(
            type(
                'MultiThreadMeta' + wrapped.__name__,
                (cls,),
                {'__doc__': wrapped.__doc__},
            )
        )


class MultiThreadWrapperWithMain(MultiThreadWrapper):
    """An extension of `MultiThreadWrapper` that exposes the wrapped instance
    corresponding to the [main_thread()](https://docs.python.org/3/library/threading.html#threading.main_thread)
    under the `.main_thread` field.

    This is useful for a falling back to a main instance when needed, but results
    in race conditions if used improperly.
    """

    def __init__(self, base):
        super().__init__(base)

    def __getattr__(self, attr):
        if attr == 'main_thread':
            return self._mtdict[main_thread().ident]
        return super().__getattr__(attr)

    def __setattr__(self, attr, value):
        if attr == 'main_thread':
            raise ValueError('Setting `main_thread` attribute is not allowed')
        else:
            super().__setattr__(attr, value)

    def __dir__(self):
        return super().__dir__() + ['main_thread']



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________


class PyNumeroEvaluationError(ArithmeticError):
    """An exception to be raised by PyNumero evaluation backends in the event
    of a failed function evaluation. This should be caught by solver interfaces
    and translated to the solver-specific evaluation error API.

    """

    pass



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Test the canonical expressions
#
import os

import pyomo.common.unittest as unittest

from pyomo.common.gsl import find_GSL
from pyomo.common.fileutils import this_file_dir
from pyomo.common.tempfiles import TempfileManager
from pyomo.environ import (
    ConcreteModel,
    Var,
    Constraint,
    Objective,
    Param,
    Block,
    ExternalFunction,
    value,
)
from ..nl_diff import load_and_compare_nl_baseline

import pyomo.repn.plugins.ampl.ampl_ as ampl_
from pyomo.repn.ampl import TextNLDebugTemplate as template

gsr = ampl_.generate_standard_repn
thisdir = this_file_dir()


class _NLWriter_suite:
    @classmethod
    def setUpClass(cls):
        cls.context = TempfileManager.new_context()
        cls.tempdir = cls.context.create_tempdir()

    @classmethod
    def tearDownClass(cls):
        cls.context.release()

    def _get_fnames(self):
        class_name, test_name = self.id().split('.')[-2:]
        prefix = test_name.replace("test_", "", 1)
        return (
            os.path.join(thisdir, prefix + ".nl.baseline"),
            os.path.join(self.tempdir, prefix + ".nl.out"),
        )

    def _compare_nl_baseline(self, baseline, testfile):
        self.assertEqual(
            *load_and_compare_nl_baseline(baseline, testfile, self._nl_version)
        )

    def test_export_nonlinear_variables(self):
        model = ConcreteModel()
        model.x = Var()
        model.y = Var()
        model.z = Var()
        model.w = Var([1, 2, 3])
        model.c = Constraint(expr=model.x == model.y**2)

        model.y.fix(3)
        test_fname = os.path.join(self.tempdir, "export_nonlinear_variables")
        model.write(
            test_fname,
            format=self._nl_version,
            io_options={'symbolic_solver_labels': True},
        )
        with open(test_fname + '.col') as f:
            names = list(map(str.strip, f.readlines()))
        assert "z" not in names  # z is not in a constraint
        assert "y" not in names  # y is fixed
        assert "x" in names
        model.write(
            test_fname,
            format=self._nl_version,
            io_options={
                'symbolic_solver_labels': True,
                'export_nonlinear_variables': [model.z],
            },
        )
        with open(test_fname + '.col') as f:
            names = list(map(str.strip, f.readlines()))
        assert "z" in names
        assert "y" not in names
        assert "x" in names
        assert "w[1]" not in names
        assert "w[2]" not in names
        assert "w[3]" not in names
        model.write(
            test_fname,
            format=self._nl_version,
            io_options={
                'symbolic_solver_labels': True,
                'export_nonlinear_variables': [model.z, model.w],
            },
        )
        with open(test_fname + '.col') as f:
            names = list(map(str.strip, f.readlines()))
        assert "z" in names
        assert "y" not in names
        assert "x" in names
        assert "w[1]" in names
        assert "w[2]" in names
        assert "w[3]" in names

        model.write(
            test_fname,
            format=self._nl_version,
            io_options={
                'symbolic_solver_labels': True,
                'export_nonlinear_variables': [model.z, model.w[2]],
            },
        )
        with open(test_fname + '.col') as f:
            names = list(map(str.strip, f.readlines()))
        assert "z" in names
        assert "y" not in names
        assert "x" in names
        assert "w[1]" not in names
        assert "w[2]" in names
        assert "w[3]" not in names

    def test_var_on_other_model(self):
        if self._nl_version != 'nl_v1':
            self.skipTest(f'test not applicable to writer {self._nl_version}')
        other = ConcreteModel()
        other.a = Var()

        model = ConcreteModel()
        model.x = Var()
        model.c = Constraint(expr=other.a + 2 * model.x <= 0)
        model.obj = Objective(expr=model.x)

        baseline_fname, test_fname = self._get_fnames()
        self.assertRaisesRegex(
            KeyError,
            "'a' is not part of the model",
            model.write,
            test_fname,
            format=self._nl_version,
        )

    def test_var_on_deactivated_block(self):
        model = ConcreteModel()
        model.x = Var()
        model.other = Block()
        model.other.a = Var()
        model.other.deactivate()
        model.c = Constraint(expr=model.other.a + 2 * model.x <= 0)
        model.obj = Objective(expr=model.x)

        baseline_fname, test_fname = self._get_fnames()
        model.write(test_fname, format=self._nl_version)
        self._compare_nl_baseline(baseline_fname, test_fname)

    def test_var_on_nonblock(self):
        if self._nl_version != 'nl_v1':
            self.skipTest(f'test not applicable to writer {self._nl_version}')

        class Foo(Block().__class__):
            def __init__(self, *args, **kwds):
                kwds.setdefault('ctype', Foo)
                super(Foo, self).__init__(*args, **kwds)

        model = ConcreteModel()
        model.x = Var()
        model.other = Foo()
        model.other.a = Var()
        model.c = Constraint(expr=model.other.a + 2 * model.x <= 0)
        model.obj = Objective(expr=model.x)

        baseline_fname, test_fname = self._get_fnames()
        self.assertRaisesRegex(
            KeyError,
            "'other.a' exists within Foo 'other'",
            model.write,
            test_fname,
            format=self._nl_version,
        )

    def _external_model(self):
        DLL = find_GSL()
        if not DLL:
            self.skipTest("Could not find the amplgsl.dll library")

        m = ConcreteModel()
        m.hypot = ExternalFunction(library=DLL, function="gsl_hypot")
        m.p = Param(initialize=1, mutable=True)
        m.x = Var(initialize=3, bounds=(1e-5, None))
        m.y = Var(initialize=3, bounds=(0, None))
        m.z = Var(initialize=1)
        m.o = Objective(expr=m.z**2 * m.hypot(m.p * m.x, m.p + m.y) ** 2)
        self.assertAlmostEqual(value(m.o), 25.0, 7)
        return m

    def test_external_expression_constant(self):
        DLL = find_GSL()
        if not DLL:
            self.skipTest("Could not find the amplgsl.dll library")

        m = ConcreteModel()
        m.y = Var(initialize=4, bounds=(0, None))
        m.hypot = ExternalFunction(library=DLL, function="gsl_hypot")
        m.o = Objective(expr=m.hypot(3, m.y))
        self.assertAlmostEqual(value(m.o), 5.0, 7)

        baseline_fname, test_fname = self._get_fnames()
        m.write(
            test_fname,
            format=self._nl_version,
            io_options={'symbolic_solver_labels': True},
        )
        self._compare_nl_baseline(baseline_fname, test_fname)

    def test_external_expression_variable(self):
        m = self._external_model()

        baseline_fname, test_fname = self._get_fnames()
        m.write(
            test_fname,
            format=self._nl_version,
            io_options={'symbolic_solver_labels': True, 'column_order': True},
        )
        self._compare_nl_baseline(baseline_fname, test_fname)

    def test_external_expression_partial_fixed(self):
        m = self._external_model()
        m.x.fix()

        baseline_fname, test_fname = self._get_fnames()
        m.write(
            test_fname,
            format=self._nl_version,
            io_options={'symbolic_solver_labels': True, 'column_order': True},
        )
        self._compare_nl_baseline(baseline_fname, test_fname)

    def test_external_expression_fixed(self):
        m = self._external_model()
        m.x.fix()
        m.y.fix()

        baseline_fname, test_fname = self._get_fnames()
        m.write(
            test_fname,
            format=self._nl_version,
            io_options={'symbolic_solver_labels': True, 'column_order': True},
        )
        self._compare_nl_baseline(baseline_fname, test_fname)

    def test_external_expression_rewrite_fixed(self):
        m = self._external_model()

        baseline_fname, test_fname = self._get_fnames()
        variable_baseline = baseline_fname.replace('rewrite_fixed', 'variable')
        m.write(
            test_fname,
            format=self._nl_version,
            io_options={'symbolic_solver_labels': True, 'column_order': True},
        )
        self._compare_nl_baseline(variable_baseline, test_fname)

        m.x.fix()
        m.write(
            test_fname,
            format=self._nl_version,
            io_options={'symbolic_solver_labels': True, 'column_order': True},
        )
        partial_baseline = baseline_fname.replace('rewrite_fixed', 'partial_fixed')
        self._compare_nl_baseline(partial_baseline, test_fname)

        m.y.fix()
        m.write(
            test_fname,
            format=self._nl_version,
            io_options={'symbolic_solver_labels': True},
        )
        fixed_baseline = baseline_fname.replace('rewrite_fixed', 'fixed')
        self._compare_nl_baseline(fixed_baseline, test_fname)

    def test_obj_con_cache(self):
        if self._nl_version != 'nl_v1':
            self.skipTest(f'test not applicable to writer {self._nl_version}')
        model = ConcreteModel()
        model.x = Var()
        model.c = Constraint(expr=model.x**2 >= 1)
        model.obj = Objective(expr=model.x**2)

        with TempfileManager.new_context() as TMP:
            nl_file = TMP.create_tempfile(suffix='.nl')
            model.write(nl_file, format=self._nl_version)
            self.assertFalse(hasattr(model, '_repn'))
            with open(nl_file) as FILE:
                nl_ref = FILE.read()

            nl_file = TMP.create_tempfile(suffix='.nl')
            model._gen_obj_repn = True
            model.write(nl_file, format=self._nl_version)
            self.assertEqual(len(model._repn), 1)
            self.assertIn(model.obj, model._repn)
            obj_repn = model._repn[model.obj]
            with open(nl_file) as FILE:
                nl_test = FILE.read()
            self.assertEqual(nl_ref, nl_test)

            nl_file = TMP.create_tempfile(suffix='.nl')
            del model._repn
            model._gen_obj_repn = None
            model._gen_con_repn = True
            model.write(nl_file, format=self._nl_version)
            self.assertEqual(len(model._repn), 1)
            self.assertIn(model.c, model._repn)
            c_repn = model._repn[model.c]
            with open(nl_file) as FILE:
                nl_test = FILE.read()
            self.assertEqual(nl_ref, nl_test)

            nl_file = TMP.create_tempfile(suffix='.nl')
            del model._repn
            model._gen_obj_repn = True
            model._gen_con_repn = True
            model.write(nl_file, format=self._nl_version)
            self.assertEqual(len(model._repn), 2)
            self.assertIn(model.obj, model._repn)
            self.assertIn(model.c, model._repn)
            obj_repn = model._repn[model.obj]
            c_repn = model._repn[model.c]
            with open(nl_file) as FILE:
                nl_test = FILE.read()
            self.assertEqual(nl_ref, nl_test)

            nl_file = TMP.create_tempfile(suffix='.nl')
            model._gen_obj_repn = None
            model._gen_con_repn = None
            model.write(nl_file, format=self._nl_version)
            self.assertEqual(len(model._repn), 2)
            self.assertIn(model.obj, model._repn)
            self.assertIn(model.c, model._repn)
            self.assertIs(obj_repn, model._repn[model.obj])
            self.assertIs(c_repn, model._repn[model.c])
            with open(nl_file) as FILE:
                nl_test = FILE.read()
            self.assertEqual(nl_ref, nl_test)

            nl_file = TMP.create_tempfile(suffix='.nl')
            model._gen_obj_repn = True
            model._gen_con_repn = True
            model.write(nl_file, format=self._nl_version)
            self.assertEqual(len(model._repn), 2)
            self.assertIn(model.obj, model._repn)
            self.assertIn(model.c, model._repn)
            self.assertIsNot(obj_repn, model._repn[model.obj])
            self.assertIsNot(c_repn, model._repn[model.c])
            obj_repn = model._repn[model.obj]
            c_repn = model._repn[model.c]
            with open(nl_file) as FILE:
                nl_test = FILE.read()
            self.assertEqual(nl_ref, nl_test)

            nl_file = TMP.create_tempfile(suffix='.nl')
            model._gen_obj_repn = False
            model._gen_con_repn = False
            try:

                def dont_call_gsr(*args, **kwargs):
                    self.fail("generate_standard_repn should not be called")

                ampl_.generate_standard_repn = dont_call_gsr
                model.write(nl_file, format=self._nl_version)
            finally:
                ampl_.generate_standard_repn = gsr
            self.assertEqual(len(model._repn), 2)
            self.assertIn(model.obj, model._repn)
            self.assertIn(model.c, model._repn)
            self.assertIs(obj_repn, model._repn[model.obj])
            self.assertIs(c_repn, model._repn[model.c])
            with open(nl_file) as FILE:
                nl_test = FILE.read()
            self.assertEqual(nl_ref, nl_test)

            # Check that repns generated by the LP wrter will be
            # processed correctly
            model._repn[model.c] = c_repn = gsr(model.c.body, quadratic=True)
            model._repn[model.obj] = obj_repn = gsr(model.obj.expr, quadratic=True)
            nl_file = TMP.create_tempfile(suffix='.nl')
            try:

                def dont_call_gsr(*args, **kwargs):
                    self.fail("generate_standard_repn should not be called")

                ampl_.generate_standard_repn = dont_call_gsr
                model.write(nl_file, format=self._nl_version)
            finally:
                ampl_.generate_standard_repn = gsr
            self.assertEqual(len(model._repn), 2)
            self.assertIn(model.obj, model._repn)
            self.assertIn(model.c, model._repn)
            self.assertIs(obj_repn, model._repn[model.obj])
            self.assertIs(c_repn, model._repn[model.c])
            with open(nl_file) as FILE:
                nl_test = FILE.read()
            self.assertEqual(nl_ref, nl_test)


class TestNLWriter_v1(_NLWriter_suite, unittest.TestCase):
    _nl_version = 'nl_v1'


class TestNLWriter_v2(_NLWriter_suite, unittest.TestCase):
    _nl_version = 'nl_v2'


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""
Methods for constructing and solving PyROS separation problems
and related objects.
"""

from itertools import product
import logging

from pyomo.common.collections import ComponentSet, ComponentMap
from pyomo.common.dependencies import numpy as np
from pyomo.core.base import Block, Constraint, maximize, Objective, value, Var
from pyomo.opt import TerminationCondition as tc
from pyomo.core.expr import (
    replace_expressions,
    identify_variables,
    identify_mutable_parameters,
)

from pyomo.contrib.pyros.solve_data import (
    DiscreteSeparationSolveCallResults,
    SeparationSolveCallResults,
    SeparationLoopResults,
    SeparationResults,
)
from pyomo.contrib.pyros.uncertainty_sets import Geometry
from pyomo.contrib.pyros.util import (
    ABS_CON_CHECK_FEAS_TOL,
    call_solver,
    check_time_limit_reached,
    get_all_first_stage_eq_cons,
    write_subproblem,
)


def add_uncertainty_set_constraints(separation_model, config):
    """
    Add to the separation model constraints restricting
    the uncertain parameter proxy variables to the user-provided
    uncertainty set. Note that inferred interval enclosures
    on the uncertain parameters are also imposed as bounds
    specified on the proxy variables.
    """
    separation_model.uncertainty = Block()
    separation_model.uncertainty.uncertain_param_indexed_var = Var(
        range(config.uncertainty_set.dim),
        initialize={
            idx: nom_val
            for idx, nom_val in enumerate(config.nominal_uncertain_param_vals)
        },
    )
    indexed_param_var = separation_model.uncertainty.uncertain_param_indexed_var
    uncertainty_quantification = config.uncertainty_set.set_as_constraint(
        uncertain_params=indexed_param_var, block=separation_model.uncertainty
    )

    # facilitate retrieval later
    _, uncertainty_cons, param_var_list, aux_vars = uncertainty_quantification
    separation_model.uncertainty.uncertain_param_var_list = param_var_list
    separation_model.uncertainty.auxiliary_var_list = aux_vars
    separation_model.uncertainty.uncertainty_cons_list = uncertainty_cons

    config.uncertainty_set._add_bounds_on_uncertain_parameters(
        uncertain_param_vars=param_var_list, global_solver=config.global_solver
    )
    if aux_vars:
        aux_var_vals = config.uncertainty_set.compute_auxiliary_uncertain_param_vals(
            point=config.nominal_uncertain_param_vals, solver=config.global_solver
        )
        for auxvar, auxval in zip(aux_vars, aux_var_vals):
            auxvar.set_value(auxval)

    # fix the effectively certain parameters
    param_val_enum_zip = enumerate(
        zip(param_var_list, config.nominal_uncertain_param_vals)
    )
    fixed_param_var_set = ComponentSet()
    for idx, (param_var, nomval) in param_val_enum_zip:
        if idx not in separation_model.effective_uncertain_dimensions:
            param_var.fix(nomval)
            fixed_param_var_set.add(param_var)

    # deactivate constraints that depend on only the
    # effective certain parameters
    separation_model.uncertainty.certain_param_var_cons = []
    for con in uncertainty_cons:
        unfixed_param_vars_in_con = (
            ComponentSet(identify_variables(con.expr)) - fixed_param_var_set
        )
        if not unfixed_param_vars_in_con:
            con.deactivate()
            separation_model.uncertainty.certain_param_var_cons.append(con)


def construct_separation_problem(model_data):
    """
    Construct the separation problem model from the fully preprocessed
    working model.

    Parameters
    ----------
    model_data : model data object
        Main model data object.

    Returns
    -------
    separation_model : ConcreteModel
        Separation problem model.
    """
    config = model_data.config
    separation_model = model_data.working_model.clone()

    # fix/deactivate all nonadjustable components
    for var in separation_model.all_nonadjustable_variables:
        var.fix()
    for fs_eqcon in get_all_first_stage_eq_cons(separation_model):
        fs_eqcon.deactivate()
    for fs_ineqcon in separation_model.first_stage.inequality_cons.values():
        fs_ineqcon.deactivate()

    # add block for the uncertainty set quantification
    add_uncertainty_set_constraints(separation_model, config)

    # the uncertain params function as decision variables
    # in the separation problems.
    # note: expression replacement is performed only for
    #       the active constraints
    uncertain_params = separation_model.uncertain_params
    uncertain_param_vars = separation_model.uncertainty.uncertain_param_var_list
    param_id_to_var_map = {
        id(param): var for param, var in zip(uncertain_params, uncertain_param_vars)
    }
    uncertain_params_set = ComponentSet(uncertain_params)
    adjustable_cons = (
        list(separation_model.second_stage.inequality_cons.values())
        + list(separation_model.second_stage.equality_cons.values())
        + list(separation_model.second_stage.decision_rule_eqns.values())
    )
    for adjcon in adjustable_cons:
        uncertain_params_in_con = (
            ComponentSet(identify_mutable_parameters(adjcon.expr))
            & uncertain_params_set
        )
        if uncertain_params_in_con:
            adjcon.set_value(
                replace_expressions(adjcon.expr, substitution_map=param_id_to_var_map)
            )

    # second-stage inequality constraint expressions
    # become maximization objectives in the separation problems
    separation_model.second_stage_ineq_con_to_obj_map = ComponentMap()
    ss_ineq_cons = separation_model.second_stage.inequality_cons.values()
    for idx, ss_ineq_con in enumerate(ss_ineq_cons):
        ss_ineq_con.deactivate()
        separation_obj = Objective(
            expr=ss_ineq_con.body - ss_ineq_con.upper, sense=maximize
        )
        separation_model.add_component(f"separation_obj_{idx}", separation_obj)
        separation_model.second_stage_ineq_con_to_obj_map[ss_ineq_con] = separation_obj
        separation_obj.deactivate()

    return separation_model


def get_sep_objective_values(separation_data, ss_ineq_cons):
    """
    Evaluate second-stage inequality constraint functions at current
    separation solution.

    Parameters
    ----------
    separation_data : SeparationProblemData
        Separation problem data.
    ss_ineq_cons : list of Constraint
        Second-stage inequality constraints to be evaluated.

    Returns
    -------
    violations : ComponentMap
        Mapping from second-stage inequality constraints
        to violation values.
    """
    config = separation_data.config
    con_to_obj_map = separation_data.separation_model.second_stage_ineq_con_to_obj_map
    violations = ComponentMap()

    for ss_ineq_con in ss_ineq_cons:
        obj = con_to_obj_map[ss_ineq_con]
        try:
            violations[ss_ineq_con] = value(obj.expr)
        except (ValueError, ArithmeticError):
            vars_in_expr_str = ",\n  ".join(
                f"{var.name}={var.value}" for var in identify_variables(obj.expr)
            )
            config.progress_logger.error(
                "PyROS encountered an exception evaluating "
                "expression of second-stage inequality constraint with name "
                f"{ss_ineq_con.name!r} (separation objective {obj.name!r}) "
                f"at variable values:\n  {vars_in_expr_str}\n"
                "Does the expression contain log(x) or 1/x functions "
                "or others with tricky domains?"
            )
            raise

    return violations


def get_argmax_sum_violations(solver_call_results_map, ss_ineq_cons_to_evaluate):
    """
    Get key of entry of `solver_call_results_map` which contains
    separation problem solution with maximal sum of second-stage
    inequality constraint violations over a specified sequence of
    second-stage inequality constraints.

    Parameters
    ----------
    solver_call_results : ComponentMap
        Mapping from second-stage inequality constraints to corresponding
        separation solver call results.
    ss_ineq_cons_to_evaluate : list of Constraints
        Second-stage inequality constraints to consider for evaluating
        maximal sum.

    Returns
    -------
    worst_ss_ineq_con : None or Constraint
        Second-stage inequality constraint corresponding to solver call
        results object containing solution with maximal sum
        of violations across all second-stage inequality constraints.
        If ``found_violation`` attribute of all value entries of
        `solver_call_results_map` is False, then `None` is
        returned, as this means
        none of the second-stage inequality constraints
        were found to be violated.
    """
    # get indices of second-stage ineq constraints
    # for which violation found
    idx_to_ss_ineq_con_map = {
        idx: ss_ineq_con for idx, ss_ineq_con in enumerate(solver_call_results_map)
    }
    idxs_of_violated_cons = [
        idx
        for idx, ss_ineq_con in idx_to_ss_ineq_con_map.items()
        if solver_call_results_map[ss_ineq_con].found_violation
    ]

    num_violated_cons = len(idxs_of_violated_cons)

    if num_violated_cons == 0:
        return None

    # assemble square matrix (2D array) of constraint violations.
    # matrix size: number of constraints for which violation was found
    # each row corresponds to a second-stage inequality constraint
    # each column corresponds to a separation problem solution
    violations_arr = np.zeros(shape=(num_violated_cons, num_violated_cons))
    idxs_product = product(
        enumerate(idxs_of_violated_cons), enumerate(idxs_of_violated_cons)
    )
    for (row_idx, viol_con_idx), (col_idx, viol_param_idx) in idxs_product:
        violations_arr[row_idx, col_idx] = max(
            0,
            (
                # violation of this row's second-stage inequality con
                # by this column's separation solution
                # if separation problems were solved globally,
                # then diagonal entries should be the largest in each row
                solver_call_results_map[
                    idx_to_ss_ineq_con_map[viol_param_idx]
                ].scaled_violations[idx_to_ss_ineq_con_map[viol_con_idx]]
            ),
        )

    worst_col_idx = np.argmax(np.sum(violations_arr, axis=0))

    return idx_to_ss_ineq_con_map[idxs_of_violated_cons[worst_col_idx]]


def solve_separation_problem(separation_data, master_data):
    """
    Solve PyROS separation problems.

    Parameters
    ----------
    separation_data : SeparationProblemData
        Separation problem data.
    master_data : MasterProblemData
        Master problem data.

    Returns
    -------
    pyros.solve_data.SeparationResults
        Separation problem solve results.
    """
    config = separation_data.config
    run_local = not config.bypass_local_separation
    run_global = config.bypass_local_separation

    uncertainty_set_is_discrete = (
        config.uncertainty_set.geometry == Geometry.DISCRETE_SCENARIOS
    )

    if run_local:
        local_separation_loop_results = perform_separation_loop(
            separation_data=separation_data,
            master_data=master_data,
            solve_globally=False,
        )
        run_global = not (
            local_separation_loop_results.found_violation
            or uncertainty_set_is_discrete
            or local_separation_loop_results.subsolver_error
            or local_separation_loop_results.time_out
            or config.bypass_global_separation
        )
    else:
        local_separation_loop_results = None

    if run_global:
        global_separation_loop_results = perform_separation_loop(
            separation_data=separation_data,
            master_data=master_data,
            solve_globally=True,
        )
    else:
        global_separation_loop_results = None

    return SeparationResults(
        local_separation_loop_results=local_separation_loop_results,
        global_separation_loop_results=global_separation_loop_results,
    )


def evaluate_violations_by_nominal_master(separation_data, master_data, ss_ineq_cons):
    """
    Evaluate violation of second-stage inequality constraints by
    variables in nominal block of most recent master
    problem.

    Returns
    -------
    nom_ss_ineq_con_violations : dict
        Mapping from second-stage inequality constraint names
        to floats equal to violations by nominal master
        problem variables.
    """
    nom_ss_ineq_con_violations = ComponentMap()
    for ss_ineq_con in ss_ineq_cons:
        nom_master_con = master_data.master_model.scenarios[0, 0].find_component(
            ss_ineq_con
        )
        nom_violation = value(nom_master_con.body - nom_master_con.upper)
        nom_ss_ineq_con_violations[ss_ineq_con] = nom_violation

    return nom_ss_ineq_con_violations


def group_ss_ineq_constraints_by_priority(separation_data):
    """
    Group model second-stage inequality constraints
    by separation priority.

    Parameters
    ----------
    separation_data : SeparationProblemData
        Separation problem data.

    Returns
    -------
    dict
        Mapping from an int to a list of second-stage
        inequality constraints
        (Constraint objects),
        for which the int is equal to the specified priority.
        Keys are sorted in descending order
        (i.e. highest priority first).
    """
    separation_data.config.progress_logger.debug(
        "Grouping second-stage inequality constraints by separation priority..."
    )

    ss_ineq_cons = separation_data.separation_model.second_stage.inequality_cons
    separation_priority_groups = dict()
    for name, ss_ineq_con in ss_ineq_cons.items():
        priority = separation_data.separation_priority_order[name]
        cons_with_same_priority = separation_priority_groups.setdefault(priority, [])
        cons_with_same_priority.append(ss_ineq_con)

    # sort separation priority groups
    numeric_priority_grp_items = [
        (priority, cons) for priority, cons in separation_priority_groups.items()
    ]
    sorted_priority_groups = {
        priority: ss_ineq_cons
        for priority, ss_ineq_cons in sorted(numeric_priority_grp_items, reverse=True)
    }

    num_priority_groups = len(sorted_priority_groups)
    separation_data.config.progress_logger.debug(
        f"Found {num_priority_groups} separation "
        f"priority group{'s' if num_priority_groups != 1 else ''}."
    )
    separation_data.config.progress_logger.debug(
        "Separation priority grouping statistics:"
    )
    separation_data.config.progress_logger.debug(
        f"  {'Priority':20s}{'# Ineq Cons':15s}"
    )
    for priority, cons in sorted_priority_groups.items():
        separation_data.config.progress_logger.debug(
            f"  {priority:<20d}{len(cons):<15d}"
        )

    return sorted_priority_groups


def get_worst_discrete_separation_solution(
    ss_ineq_con, config, ss_ineq_cons_to_evaluate, discrete_solve_results
):
    """
    Determine separation solution (and therefore worst-case
    uncertain parameter realization) with maximum violation
    of specified second-stage inequality constraint.

    Parameters
    ----------
    ss_ineq_con : Constraint
        Second-stage inequality constraint of interest.
    config : ConfigDict
        User-specified PyROS solver settings.
    ss_ineq_cons_to_evaluate : list of Constraint
        Second-stage inequality constraints for which to report
        violations by separation solution.
    discrete_solve_results : DiscreteSeparationSolveCallResults
        Separation problem solutions corresponding to the
        uncertain parameter scenarios listed in
        ``config.uncertainty_set.scenarios``.

    Returns
    -------
    SeparationSolveCallResult
        Solver call result for second-stage inequality constraint of interest.
    """
    # violation of specified second-stage inequality
    # constraint by separation
    # problem solutions for all scenarios
    # scenarios with subsolver errors are replaced with nan
    violations_of_ss_ineq_con = [
        (
            solve_call_res.scaled_violations[ss_ineq_con]
            if not solve_call_res.subsolver_error
            else np.nan
        )
        for solve_call_res in discrete_solve_results.solver_call_results.values()
    ]

    list_of_scenario_idxs = list(discrete_solve_results.solver_call_results.keys())

    # determine separation solution for which scaled violation of this
    # second-stage inequality constraint is the worst
    worst_case_res = discrete_solve_results.solver_call_results[
        list_of_scenario_idxs[np.nanargmax(violations_of_ss_ineq_con)]
    ]
    worst_case_violation = np.nanmax(violations_of_ss_ineq_con)
    assert worst_case_violation in worst_case_res.scaled_violations.values()

    # evaluate violations for specified second-stage inequality constraints
    eval_ss_ineq_con_scaled_violations = ComponentMap(
        (ss_ineq_con, worst_case_res.scaled_violations[ss_ineq_con])
        for ss_ineq_con in ss_ineq_cons_to_evaluate
    )

    # discrete separation solutions were obtained by optimizing
    # just one second-stage inequality constraint, as an efficiency.
    # if the constraint passed to this routine is the same as the
    # constraint used to obtain the solutions, then we bundle
    # the separation solve call results into a single list.
    # otherwise, we return an empty list, as we did not need to call
    # subsolvers for the other second-stage inequality constraints
    is_optimized_ss_ineq_con = (
        ss_ineq_con is discrete_solve_results.second_stage_ineq_con
    )
    if is_optimized_ss_ineq_con:
        results_list = [
            res
            for solve_call_results in discrete_solve_results.solver_call_results.values()
            for res in solve_call_results.results_list
        ]
    else:
        results_list = []

    # check if there were any failed scenarios for subsolver_error
    # if there are failed scenarios, subsolver error triggers for all ineq
    if any(np.isnan(violations_of_ss_ineq_con)):
        subsolver_error_flag = True
    else:
        subsolver_error_flag = False

    return SeparationSolveCallResults(
        solved_globally=worst_case_res.solved_globally,
        results_list=results_list,
        scaled_violations=eval_ss_ineq_con_scaled_violations,
        violating_param_realization=worst_case_res.violating_param_realization,
        variable_values=worst_case_res.variable_values,
        found_violation=(worst_case_violation > config.robust_feasibility_tolerance),
        time_out=False,
        subsolver_error=subsolver_error_flag,
        discrete_set_scenario_index=worst_case_res.discrete_set_scenario_index,
    )


def get_con_name_repr(separation_model, con, with_obj_name=True):
    """
    Get string representation of second-stage inequality constraint
    and the objective to which it has been mapped.

    Parameters
    ----------
    separation_model : ConcreteModel
        Separation model.
    con : ScalarConstraint or ConstraintData
        Constraint for which to get the representation.
    with_obj_name : bool, optional
        Include name of separation model objective to which
        constraint is mapped. Applicable only to second-stage inequality
        constraints of the separation problem.

    Returns
    -------
    str
        Constraint name representation.
    """
    qual_str = ""
    if with_obj_name:
        objectives_map = separation_model.second_stage_ineq_con_to_obj_map
        separation_obj = objectives_map[con]
        qual_str = f" (mapped to objective {separation_obj.name!r})"

    return f"{con.index()!r}{qual_str}"


def perform_separation_loop(separation_data, master_data, solve_globally):
    """
    Loop through, and solve, PyROS separation problems to
    desired optimality condition.

    Parameters
    ----------
    separation_data : SeparationProblemData
        Separation problem data.
    master_data : MasterProblemData
        Master problem data.
    solve_globally : bool
        True to solve separation problems globally,
        False to solve separation problems locally.

    Returns
    -------
    pyros.solve_data.SeparationLoopResults
        Separation problem solve results.
    """
    config = separation_data.config
    all_ss_ineq_constraints = list(
        separation_data.separation_model.second_stage.inequality_cons.values()
    )
    if not all_ss_ineq_constraints:
        # robustness certified: no separation problems to solve
        return SeparationLoopResults(
            solver_call_results=ComponentMap(),
            solved_globally=solve_globally,
            worst_case_ss_ineq_con=None,
        )

    # needed for normalizing separation solution constraint violations
    separation_data.nom_ss_ineq_con_violations = evaluate_violations_by_nominal_master(
        separation_data=separation_data,
        master_data=master_data,
        ss_ineq_cons=all_ss_ineq_constraints,
    )
    sorted_priority_groups = separation_data.separation_priority_groups
    uncertainty_set_is_discrete = (
        config.uncertainty_set.geometry == Geometry.DISCRETE_SCENARIOS
    )

    if uncertainty_set_is_discrete:
        all_scenarios_exhausted = len(separation_data.idxs_of_master_scenarios) == len(
            config.uncertainty_set.scenarios
        )
        if all_scenarios_exhausted:
            # robustness certified: entire uncertainty set already
            # accounted for in master
            return SeparationLoopResults(
                solver_call_results=ComponentMap(),
                solved_globally=solve_globally,
                worst_case_ss_ineq_con=None,
                all_discrete_scenarios_exhausted=True,
            )

        ss_ineq_con_to_maximize = sorted_priority_groups[
            max(sorted_priority_groups.keys())
        ][0]

        # efficiency: evaluate all separation problem solutions in
        # advance of entering loop
        discrete_sep_results = discrete_solve(
            separation_data=separation_data,
            master_data=master_data,
            solve_globally=solve_globally,
            ss_ineq_con_to_maximize=ss_ineq_con_to_maximize,
            ss_ineq_cons_to_evaluate=all_ss_ineq_constraints,
        )

        termination_not_ok = (
            discrete_sep_results.time_out or discrete_sep_results.subsolver_error
        )
        if termination_not_ok:
            single_solver_call_res = ComponentMap()
            results_list = [
                res
                for solve_call_results in discrete_sep_results.solver_call_results.values()
                for res in solve_call_results.results_list
            ]
            single_solver_call_res[ss_ineq_con_to_maximize] = (
                # not the neatest assembly,
                # but should maintain accuracy of total solve times
                # and overall outcome
                SeparationSolveCallResults(
                    solved_globally=solve_globally,
                    results_list=results_list,
                    time_out=discrete_sep_results.time_out,
                    subsolver_error=discrete_sep_results.subsolver_error,
                )
            )
            return SeparationLoopResults(
                solver_call_results=single_solver_call_res,
                solved_globally=solve_globally,
                worst_case_ss_ineq_con=None,
            )

    all_solve_call_results = ComponentMap()
    priority_groups_enum = enumerate(sorted_priority_groups.items())
    solve_adverb = "Globally" if solve_globally else "Locally"
    for group_idx, (priority, ss_ineq_constraints) in priority_groups_enum:
        priority_group_solve_call_results = ComponentMap()

        for idx, ss_ineq_con in enumerate(ss_ineq_constraints):
            # log progress of separation loop
            config.progress_logger.debug(
                f"{solve_adverb} separating second-stage inequality constraint "
                f"{get_con_name_repr(separation_data.separation_model, ss_ineq_con)} "
                f"(priority {priority}, priority group {group_idx + 1} of "
                f"{len(sorted_priority_groups)}, "
                f"constraint {idx + 1} of {len(ss_ineq_constraints)} "
                "in priority group, "
                f"{len(all_solve_call_results) + idx + 1} of "
                f"{len(all_ss_ineq_constraints)} total)"
            )

            # solve separation problem for
            # this second-stage inequality constraint
            if uncertainty_set_is_discrete:
                solve_call_results = get_worst_discrete_separation_solution(
                    ss_ineq_con=ss_ineq_con,
                    config=config,
                    ss_ineq_cons_to_evaluate=all_ss_ineq_constraints,
                    discrete_solve_results=discrete_sep_results,
                )
            else:
                solve_call_results = solver_call_separation(
                    separation_data=separation_data,
                    master_data=master_data,
                    solve_globally=solve_globally,
                    ss_ineq_con_to_maximize=ss_ineq_con,
                    ss_ineq_cons_to_evaluate=all_ss_ineq_constraints,
                )

            priority_group_solve_call_results[ss_ineq_con] = solve_call_results

            if solve_call_results.time_out:
                all_solve_call_results.update(priority_group_solve_call_results)
                return SeparationLoopResults(
                    solver_call_results=all_solve_call_results,
                    solved_globally=solve_globally,
                    worst_case_ss_ineq_con=None,
                )
            elif not solve_call_results.subsolver_error:
                config.progress_logger.debug("Separation successful. Results: ")
                config.progress_logger.debug(
                    " Scaled violation: "
                    f"{solve_call_results.scaled_violations[ss_ineq_con]}"
                )
                config.progress_logger.debug(
                    " Worst-case violating realization: "
                    f"{solve_call_results.violating_param_realization}"
                )
                config.progress_logger.debug(
                    f" Is constraint violated: {solve_call_results.found_violation} "
                    f"(compared to tolerance {config.robust_feasibility_tolerance})"
                )

            # provide message that PyROS will attempt to find a violation and move
            # to the next iteration even after subsolver error
            if solve_call_results.subsolver_error:
                config.progress_logger.warning(
                    "PyROS is attempting to recover and will continue to "
                    "the next iteration if a constraint violation is found."
                )

        all_solve_call_results.update(priority_group_solve_call_results)

        config.progress_logger.debug(
            f"Done separating all constraints of priority {priority} "
            f"(group {group_idx + 1} of {len(sorted_priority_groups)})"
        )

        # there may be multiple separation problem solutions
        # found to have violated a second-stage inequality constraint.
        # we choose just one for master problem of next iteration
        worst_case_ss_ineq_con = get_argmax_sum_violations(
            solver_call_results_map=all_solve_call_results,
            ss_ineq_cons_to_evaluate=ss_ineq_constraints,
        )
        if worst_case_ss_ineq_con is not None:
            # take note of chosen separation solution
            worst_case_res = all_solve_call_results[worst_case_ss_ineq_con]
            if uncertainty_set_is_discrete:
                separation_data.idxs_of_master_scenarios.append(
                    worst_case_res.discrete_set_scenario_index
                )

            # # auxiliary log messages
            violated_con_names = "\n ".join(
                get_con_name_repr(separation_data.separation_model, con)
                for con, res in all_solve_call_results.items()
                if res.found_violation
            )
            config.progress_logger.debug(
                f"Violated constraints:\n {violated_con_names} "
            )
            config.progress_logger.debug(
                "Worst-case constraint: "
                f"{get_con_name_repr(separation_data.separation_model, worst_case_ss_ineq_con)} "
                "under realization "
                f"{worst_case_res.violating_param_realization}."
            )
            config.progress_logger.debug(
                f"Maximal scaled violation "
                f"{worst_case_res.scaled_violations[worst_case_ss_ineq_con]} "
                "from this constraint "
                "exceeds the robust feasibility tolerance "
                f"{config.robust_feasibility_tolerance}"
            )

            # violating separation problem solution now chosen.
            # exit loop
            break
        else:
            config.progress_logger.debug(
                "No violated second-stage inequality constraints found."
            )

    return SeparationLoopResults(
        solver_call_results=all_solve_call_results,
        solved_globally=solve_globally,
        worst_case_ss_ineq_con=worst_case_ss_ineq_con,
    )


def evaluate_ss_ineq_con_violations(
    separation_data, ss_ineq_con_to_maximize, ss_ineq_cons_to_evaluate
):
    """
    Evaluate the inequality constraint function violations
    of the current separation model solution, and store the
    results in a given `SeparationResult` object.
    Also, determine whether the separation solution violates
    the inequality constraint whose body is the model's
    active objective.

    Parameters
    ----------
    separation_data : SeparationProblemData
        Object containing the separation model.
    ss_ineq_con_to_maximize : ConstraintData
        Second-stage inequality constraint
        to which the current solution is mapped.
    ss_ineq_cons_to_evaluate : list of Constraint
        Second-stage inequality constraints whose expressions are to
        be evaluated at the current separation problem
        solution.
        Exactly one of these constraints should be mapped
        to an active Objective in the separation model.

    Returns
    -------
    violating_param_realization : list of float
        Uncertain parameter realization corresponding to maximum
        constraint violation.
    scaled_violations : ComponentMap
        Mapping from second-stage inequality constraints to be evaluated
        to their violations by the separation problem solution.
    constraint_violated : bool
        True if second-stage inequality constraint mapped to active
        separation model Objective is violated (beyond tolerance),
        False otherwise

    Raises
    ------
    ValueError
        If `ss_ineq_cons_to_evaluate` does not contain exactly
        1 entry which can be mapped to an active Objective
        of ``model_data.separation_model``.
    """
    config = separation_data.config

    # parameter realization for current separation problem solution
    uncertain_param_vars = (
        separation_data.separation_model.uncertainty.uncertain_param_var_list
    )
    violating_param_realization = list(
        param_var.value for param_var in uncertain_param_vars
    )

    # evaluate violations for all second-stage inequality
    # constraints provided
    violations_by_sep_solution = get_sep_objective_values(
        separation_data=separation_data, ss_ineq_cons=ss_ineq_cons_to_evaluate
    )

    # normalize constraint violation: i.e. divide by
    # absolute value of constraint expression evaluated at
    # nominal master solution (if expression value is large enough)
    scaled_violations = ComponentMap()
    for ss_ineq_con, sep_sol_violation in violations_by_sep_solution.items():
        scaled_violation = sep_sol_violation / max(
            1, abs(separation_data.nom_ss_ineq_con_violations[ss_ineq_con])
        )
        scaled_violations[ss_ineq_con] = scaled_violation
        if ss_ineq_con is ss_ineq_con_to_maximize:
            scaled_active_obj_violation = scaled_violation

    constraint_violated = (
        scaled_active_obj_violation > config.robust_feasibility_tolerance
    )

    return (violating_param_realization, scaled_violations, constraint_violated)


def initialize_separation(ss_ineq_con_to_maximize, separation_data, master_data):
    """
    Initialize separation problem variables using the solution
    to the most recent master problem.

    Parameters
    ----------
    ss_ineq_con_to_maximize : ConstraintData
        Second-stage inequality constraint
        whose violation is to be maximized
        for the separation problem of interest.
    separation_data : SeparationProblemData
        Separation problem data.
    master_data : MasterProblemData
        Master problem data.

    Note
    ----
    The point to which the separation model is initialized should,
    in general, be feasible, provided the set does not have a
    discrete geometry (as there is no master model block corresponding
    to any of the remaining discrete scenarios against which we
    separate).
    """
    config = separation_data.config
    master_model = master_data.master_model
    sep_model = separation_data.separation_model

    def eval_master_violation(scenario_idx):
        """
        Evaluate violation of `ss_ineq_con` by variables of
        specified master block.
        """
        master_con = master_model.scenarios[scenario_idx].find_component(
            ss_ineq_con_to_maximize
        )
        return value(master_con)

    # initialize from master block with max violation of the
    # second-stage ineq constraint of interest. Gives the best known
    # feasible solution (for case of non-discrete uncertainty sets).
    worst_master_block_idx = max(
        master_model.scenarios.keys(), key=eval_master_violation
    )
    worst_case_master_blk = master_model.scenarios[worst_master_block_idx]
    for sep_var in sep_model.all_variables:
        master_var = worst_case_master_blk.find_component(sep_var)
        sep_var.set_value(value(master_var, exception=False))

    # for discrete uncertainty sets, the uncertain parameters
    # have already been addressed
    if config.uncertainty_set.geometry != Geometry.DISCRETE_SCENARIOS:
        param_vars = sep_model.uncertainty.uncertain_param_var_list
        param_values = separation_data.points_added_to_master[worst_master_block_idx]
        for param_var, val in zip(param_vars, param_values):
            param_var.set_value(val)

        aux_param_vars = sep_model.uncertainty.auxiliary_var_list
        aux_param_values = separation_data.auxiliary_values_for_master_points[
            worst_master_block_idx
        ]
        for aux_param_var, aux_val in zip(aux_param_vars, aux_param_values):
            aux_param_var.set_value(aux_val)

    # confirm the initial point is feasible for cases where
    # we expect it to be (i.e. non-discrete uncertainty sets).
    # otherwise, log the violated constraints
    if config.progress_logger.isEnabledFor(logging.DEBUG):
        tol = ABS_CON_CHECK_FEAS_TOL
        ss_ineq_con_name_repr = get_con_name_repr(
            separation_model=sep_model, con=ss_ineq_con_to_maximize, with_obj_name=True
        )
        uncertainty_set_is_discrete = (
            config.uncertainty_set.geometry is Geometry.DISCRETE_SCENARIOS
        )
        for con in sep_model.component_data_objects(Constraint, active=True):
            lslack, uslack = con.lslack(), con.uslack()
            if (lslack < -tol or uslack < -tol) and not uncertainty_set_is_discrete:
                config.progress_logger.debug(
                    f"Initial point for separation of second-stage ineq constraint "
                    f"{ss_ineq_con_name_repr} violates the model constraint "
                    f"{con.name!r} by more than {tol} ({lslack=}, {uslack=})"
                )

    for con in sep_model.uncertainty.certain_param_var_cons:
        trivially_infeasible = (
            con.lslack() < -ABS_CON_CHECK_FEAS_TOL
            or con.uslack() < -ABS_CON_CHECK_FEAS_TOL
        )
        if trivially_infeasible:
            # this should never happen in the context of a full solve,
            # since the certain parameters should be at their
            # nominal values, and the nominal point was already
            # confirmed to be a member of the set
            config.progress_logger.error(
                f"Uncertainty set "
                f"(type {type(config.uncertainty_set).__name__}) "
                f"constraint {con.name!r}, with expression {con.expr}, "
                "is trivially infeasible at the parameter realization "
                f"{config.nominal_uncertain_param_vals}. "
                f"Check implementation of "
                f"{config.uncertainty_set.set_as_constraint.__name__}."
            )
            raise ValueError(
                f"Trivial infeasibility detected in the uncertainty set "
                f"(type {type(config.uncertainty_set).__name__}) constraints."
            )


locally_acceptable = {tc.optimal, tc.locallyOptimal, tc.globallyOptimal}
globally_acceptable = {tc.optimal, tc.globallyOptimal}


def solver_call_separation(
    separation_data,
    master_data,
    solve_globally,
    ss_ineq_con_to_maximize,
    ss_ineq_cons_to_evaluate,
):
    """
    Invoke subordinate solver(s) on separation problem.

    Parameters
    ----------
    separation_data : SeparationProblemData
        Separation problem data.
    master_data : MasterProblemData
        Master problem data.
    solve_globally : bool
        True to solve separation problems globally,
        False to solve locally.
    ss_ineq_con_to_maximize : Constraint
        Second-stage inequality constraint
        for which to solve separation problem.
        Informs the objective (constraint violation) to maximize.
    ss_ineq_cons_to_evaluate : list of Constraint
        Second-stage inequality constraints whose expressions are to be
        evaluated at the separation problem solution
        obtained.

    Returns
    -------
    solve_call_results : pyros.solve_data.SeparationSolveCallResults
        Solve results for separation problem of interest.
    """
    config = separation_data.config
    # prepare the problem
    separation_model = separation_data.separation_model
    objectives_map = separation_data.separation_model.second_stage_ineq_con_to_obj_map
    separation_obj = objectives_map[ss_ineq_con_to_maximize]
    initialize_separation(ss_ineq_con_to_maximize, separation_data, master_data)
    separation_obj.activate()

    # get name (index) of constraint for loggers
    con_name_repr = get_con_name_repr(
        separation_model=separation_model,
        con=ss_ineq_con_to_maximize,
        with_obj_name=True,
    )

    # keep track of solver statuses for output logging
    solve_mode = "global" if solve_globally else "local"
    solver_status_dict = {}
    if solve_globally:
        solvers = [config.global_solver] + config.backup_global_solvers
    else:
        solvers = [config.local_solver] + config.backup_local_solvers
    solve_mode_adverb = "globally" if solve_globally else "locally"

    solve_call_results = SeparationSolveCallResults(
        solved_globally=solve_globally,
        time_out=False,
        results_list=[],
        found_violation=False,
        subsolver_error=False,
    )
    for idx, opt in enumerate(solvers):
        if idx > 0:
            config.progress_logger.debug(
                f"Invoking backup solver {opt!r} "
                f"(solver {idx + 1} of {len(solvers)}) for {solve_mode} "
                f"separation of second-stage inequality constraint {con_name_repr} "
                f"in iteration {separation_data.iteration}."
            )
            # TODO: confirm this is sufficient for tracking
            #       discrete separation backup solver usage
            solve_call_results.backup_solver_used = True
        results = call_solver(
            model=separation_model,
            solver=opt,
            config=config,
            timing_obj=separation_data.timing,
            timer_name=f"main.{solve_mode}_separation",
            err_msg=(
                f"Optimizer {repr(opt)} ({idx + 1} of {len(solvers)}) "
                f"encountered exception attempting "
                f"to {solve_mode_adverb} solve separation problem for constraint "
                f"{con_name_repr} in iteration {separation_data.iteration}."
            ),
        )

        # record termination condition for this particular solver
        solver_status_dict[str(opt)] = results.solver.termination_condition
        solve_call_results.results_list.append(results)

        # has PyROS time limit been reached?
        if check_time_limit_reached(separation_data.timing, config):
            solve_call_results.time_out = True
            separation_obj.deactivate()
            return solve_call_results

        # if separation problem solved to optimality, record results
        # and exit
        acceptable_conditions = (
            globally_acceptable if solve_globally else locally_acceptable
        )
        optimal_termination = solve_call_results.termination_acceptable(
            acceptable_conditions
        )
        if optimal_termination:
            separation_model.solutions.load_from(results)

            # record second-stage and state variable values
            solve_call_results.variable_values = ComponentMap()
            for var in separation_model.all_adjustable_variables:
                solve_call_results.variable_values[var] = value(var)

            # record uncertain parameter realization
            #   and constraint violations
            (
                solve_call_results.violating_param_realization,
                solve_call_results.scaled_violations,
                solve_call_results.found_violation,
            ) = evaluate_ss_ineq_con_violations(
                separation_data=separation_data,
                ss_ineq_con_to_maximize=ss_ineq_con_to_maximize,
                ss_ineq_cons_to_evaluate=ss_ineq_cons_to_evaluate,
            )
            solve_call_results.auxiliary_param_values = [
                auxvar.value
                for auxvar in separation_model.uncertainty.auxiliary_var_list
            ]

            separation_obj.deactivate()

            return solve_call_results
        else:
            config.progress_logger.debug(
                f"Solver {opt} ({idx + 1} of {len(solvers)}) "
                f"failed for {solve_mode} separation of second-stage inequality "
                f"constraint {con_name_repr} in iteration "
                f"{separation_data.iteration}. Termination condition: "
                f"{results.solver.termination_condition!r}."
            )
            config.progress_logger.debug(f"Results:\n{results.solver}")

    # All subordinate solvers failed to optimize model to appropriate
    # termination condition. PyROS will terminate with subsolver
    # error. At this point, export model if desired
    solve_call_results.subsolver_error = True
    solve_call_results.message = (
        "Could not successfully solve separation problem of iteration "
        f"{separation_data.iteration} "
        f"for second-stage inequality constraint {con_name_repr} with any of the "
        f"provided subordinate {solve_mode} optimizers. "
        f"(Termination statuses: "
        f"{[str(term_cond) for term_cond in solver_status_dict.values()]}.)"
    )
    config.progress_logger.warning(solve_call_results.message)

    if config.keepfiles and config.subproblem_file_directory is not None:
        write_subproblem(
            model=separation_model,
            fname=(
                f"{config.uncertainty_set.type}_{separation_model.name}"
                f"_separation_{separation_data.iteration}"
                f"_obj_{separation_obj.name}"
            ),
            config=config,
        )

    separation_obj.deactivate()

    return solve_call_results


def discrete_solve(
    separation_data,
    master_data,
    solve_globally,
    ss_ineq_con_to_maximize,
    ss_ineq_cons_to_evaluate,
):
    """
    Obtain separation problem solution for each scenario
    of the uncertainty set not already added to the most
    recent master problem.

    Parameters
    ----------
    separation_data : SeparationProblemData
        Separation problem data.
    master_data : MasterProblemData
        Master problem data.
    solver : solver type
        Primary subordinate optimizer with which to solve
        the model.
    solve_globally : bool
        Is separation problem to be solved globally.
    ss_ineq_con_to_maximize : Constraint
        Second-stage inequality constraint for which to solve separation
        problem.
    ss_ineq_cons_to_evaluate : list of Constraint
        Secnod-stage inequality constraints whose expressions are to be
        evaluated at the each of separation problem solutions
        obtained.

    Returns
    -------
    discrete_separation_results : DiscreteSeparationSolveCallResults
        Separation solver call results on second-stage inequality constraint
        of interest for every scenario considered.

    Notes
    -----
    Since we assume that models passed to PyROS are such that the DOF
    variables and uncertain parameter values uniquely define the state
    variables, this method need be only be invoked once per separation
    loop. Subject to our assumption, the choice of objective
    (``ss_ineq_con_to_maximize``) should not affect the solutions returned
    beyond subsolver tolerances.
    For other second-stage inequality constraints, the
    optimal separation problem solution can then be evaluated by simple
    enumeration of the solutions returned by this function, since for
    discrete uncertainty sets, the number of feasible separation
    solutions is, under our assumption, merely equal to the number
    of scenarios in the uncertainty set.
    """
    config = separation_data.config

    uncertain_param_vars = list(
        separation_data.separation_model.uncertainty.uncertain_param_var_list
    )

    # skip scenarios already added to most recent master problem
    master_scenario_idxs = separation_data.idxs_of_master_scenarios
    scenario_idxs_to_separate = [
        idx
        for idx, _ in enumerate(config.uncertainty_set.scenarios)
        if idx not in master_scenario_idxs
    ]

    solve_call_results_dict = {}
    for idx, scenario_idx in enumerate(scenario_idxs_to_separate):
        # fix uncertain parameters to scenario value
        # hence, no need to activate uncertainty set constraints
        scenario = config.uncertainty_set.scenarios[scenario_idx]
        for param, coord_val in zip(uncertain_param_vars, scenario):
            param.fix(coord_val)

        # debug statement for solving square problem for each scenario
        config.progress_logger.debug(
            f"Attempting to solve square problem for discrete scenario {scenario}"
            f" ({idx + 1} of {len(scenario_idxs_to_separate)} total)"
        )

        # obtain separation problem solution
        solve_call_results = solver_call_separation(
            separation_data=separation_data,
            master_data=master_data,
            solve_globally=solve_globally,
            ss_ineq_con_to_maximize=ss_ineq_con_to_maximize,
            ss_ineq_cons_to_evaluate=ss_ineq_cons_to_evaluate,
        )
        solve_call_results.discrete_set_scenario_index = scenario_idx
        solve_call_results_dict[scenario_idx] = solve_call_results

        # halt at first encounter of unacceptable termination
        termination_not_ok = solve_call_results.time_out
        if termination_not_ok:
            break

        # report any subsolver errors, but continue
        if solve_call_results.subsolver_error:
            config.progress_logger.warning(
                f"All solvers failed to solve discrete scenario {scenario_idx}: "
                f"{scenario}"
            )

    return DiscreteSeparationSolveCallResults(
        solved_globally=solve_globally,
        solver_call_results=solve_call_results_dict,
        second_stage_ineq_con=ss_ineq_con_to_maximize,
    )


class SeparationProblemData:
    """
    Container for objects related to the PyROS separation problem.

    Parameters
    ----------
    model_data : ModelData
        PyROS model data object, equipped with the
        fully preprocessed working model.

    Attributes
    ----------
    separation_model : BlockData
        Separation problem model object.
    timing : TimingData
        Main timer for the current problem being solved.
    config : ConfigDict
        PyROS solver options.
    separation_priority_order : dict
        Standardized/preprocessed mapping from names of the
        second-stage inequality constraint objects to integers
        specifying their priorities.
    iteration : int
        Index of the current PyROS cutting set iteration.
    points_added_to_master : dict
        Maps each scenario index (2-tuple of ints) of the
        master problem model object to the corresponding
        uncertain parameter realization.
    auxiliary_values_for_master_points : dict
        Maps each scenario index (2-tuple of ints) of the
        master problem model object to the auxiliary parameter
        values corresponding to the associated uncertain parameter
        realization.
    idxs_of_master_scenarios : None or list of int
        If ``config.uncertainty_set`` is of type
        :class:`~pyomo.contrib.pyros.uncertainty_sets.DiscreteScenarioSet`,
        then this attribute is a list
        of ints, each entry of which is a list index for
        an entry in the ``scenarios`` attribute of the
        uncertainty set. Otherwise, this attribute is set to None.
    """

    def __init__(self, model_data):
        """Initialize self (see class docstring)."""
        self.separation_model = construct_separation_problem(model_data)
        self.timing = model_data.timing
        self.separation_priority_order = model_data.separation_priority_order.copy()
        self.iteration = 0

        config = model_data.config
        self.config = config
        self.points_added_to_master = {(0, 0): config.nominal_uncertain_param_vals}
        self.auxiliary_values_for_master_points = {
            (0, 0): [
                # auxiliary variable values for nominal point have already
                # been computed and loaded into separation model
                aux_var.value
                for aux_var in self.separation_model.uncertainty.auxiliary_var_list
            ]
        }

        if config.uncertainty_set.geometry == Geometry.DISCRETE_SCENARIOS:
            self.idxs_of_master_scenarios = [
                config.uncertainty_set.scenarios.index(
                    tuple(config.nominal_uncertain_param_vals)
                )
            ]
        else:
            self.idxs_of_master_scenarios = None

        self.separation_priority_groups = group_ss_ineq_constraints_by_priority(self)

    def solve_separation(self, master_data):
        """
        Solve the separation problem.
        """
        return solve_separation_problem(self, master_data)
